from collections.abc import Callable

from esphome import automation
import esphome.codegen as cg
from esphome.components import audio
import esphome.config_validation as cv
from esphome.const import (
    CONF_ENTITY_CATEGORY,
    CONF_FORMAT,
    CONF_ICON,
    CONF_ID,
    CONF_NUM_CHANNELS,
    CONF_ON_IDLE,
    CONF_ON_STATE,
    CONF_ON_TURN_OFF,
    CONF_ON_TURN_ON,
    CONF_SAMPLE_RATE,
    CONF_VOLUME,
)
from esphome.core import CORE
from esphome.core.entity_helpers import (
    entity_duplicate_validator,
    inherit_property_from,
    queue_entity_register,
    setup_entity,
)
from esphome.coroutine import CoroPriority, coroutine_with_priority
from esphome.cpp_generator import MockObj, MockObjClass
from esphome.types import ConfigType

CODEOWNERS = ["@jesserockz"]

IS_PLATFORM_COMPONENT = True

media_player_ns = cg.esphome_ns.namespace("media_player")

MediaPlayer = media_player_ns.class_("MediaPlayer")

MediaPlayerSupportedFormat = media_player_ns.struct("MediaPlayerSupportedFormat")

MediaPlayerFormatPurpose = media_player_ns.enum(
    "MediaPlayerFormatPurpose", is_class=True
)
MEDIA_PLAYER_FORMAT_PURPOSE_ENUM = {
    "default": MediaPlayerFormatPurpose.PURPOSE_DEFAULT,
    "announcement": MediaPlayerFormatPurpose.PURPOSE_ANNOUNCEMENT,
}

# Public API for external components. Do not remove.
FORMAT_MAPPING = {
    "FLAC": "flac",
    "MP3": "mp3",
    "OPUS": "opus",
    "WAV": "wav",
}


def build_supported_format_struct(
    format_config: ConfigType, purpose: MockObj
) -> cg.StructInitializer:
    """Build a MediaPlayerSupportedFormat struct from a format config and purpose.

    Public API for external components. Do not remove.
    """
    args = [
        MediaPlayerSupportedFormat,
        ("format", FORMAT_MAPPING[format_config[CONF_FORMAT]]),
        ("sample_rate", format_config[CONF_SAMPLE_RATE]),
        ("num_channels", format_config[CONF_NUM_CHANNELS]),
        ("purpose", purpose),
    ]

    # Omit sample_bytes for MP3: ffmpeg transcoding in Home Assistant fails
    # if the number of bytes per sample is specified for MP3.
    if format_config[CONF_FORMAT] != "MP3":
        args.append(("sample_bytes", 2))

    return cg.StructInitializer(*args)


def validate_preferred_format(
    component_name: str, audio_device_key: str
) -> Callable[[ConfigType], ConfigType]:
    """Return a validator that inherits audio device settings and validates format constraints.

    Public API for external components. Do not remove.
    """

    def validator(config: ConfigType) -> ConfigType:
        # Inherit settings from audio device if not manually set
        inherit_property_from(CONF_NUM_CHANNELS, audio_device_key)(config)
        inherit_property_from(CONF_SAMPLE_RATE, audio_device_key)(config)

        # Opus only supports 48 kHz
        if config.get(CONF_FORMAT) == "OPUS" and config.get(CONF_SAMPLE_RATE) != 48000:
            raise cv.Invalid("Opus only supports a sample rate of 48000 Hz")

        # Validate the settings are compatible with the audio device
        audio.final_validate_audio_schema(
            component_name,
            audio_device=audio_device_key,
            bits_per_sample=16,
            channels=config.get(CONF_NUM_CHANNELS),
            sample_rate=config.get(CONF_SAMPLE_RATE),
        )(config)

        return config

    return validator


def request_codecs_for_format_configs(
    config: ConfigType, format_config_keys: list[str]
) -> None:
    """Scan format configs for configured formats and request the needed codec support.

    If any config uses "NONE" (accepts any format), all codecs are requested.

    Public API for external components. Do not remove.
    """
    needed_formats: set[str] = set()
    need_all = False

    for key in format_config_keys:
        if format_config := config.get(key):
            fmt = format_config[CONF_FORMAT]
            if fmt == "NONE":
                need_all = True
            else:
                needed_formats.add(fmt)

    if need_all:
        audio.request_flac_support()
        audio.request_mp3_support()
        audio.request_opus_support()
        audio.request_wav_support()
    else:
        if "FLAC" in needed_formats:
            audio.request_flac_support()
        if "MP3" in needed_formats:
            audio.request_mp3_support()
        if "OPUS" in needed_formats:
            audio.request_opus_support()
        if "WAV" in needed_formats:
            audio.request_wav_support()


# Local config key constants
CONF_ANNOUNCEMENT = "announcement"
CONF_ON_PLAY = "on_play"
CONF_ON_PAUSE = "on_pause"
CONF_ON_ANNOUNCEMENT = "on_announcement"
CONF_MEDIA_URL = "media_url"

# Command actions that all share the same schema and codegen handler
_COMMAND_ACTIONS = [
    "play",
    "pause",
    "stop",
    "toggle",
    "volume_up",
    "volume_down",
    "turn_on",
    "turn_off",
    "next",
    "previous",
    "mute",
    "unmute",
    "repeat_off",
    "repeat_one",
    "repeat_all",
    "shuffle",
    "unshuffle",
    "group_join",
    "clear_playlist",
]

StateAnyForwarder = media_player_ns.class_("StateAnyForwarder")
StateEnterForwarder = media_player_ns.class_("StateEnterForwarder")
MediaPlayerState = media_player_ns.enum("MediaPlayerState")

# State triggers: (config_key, state enum or None for any-state)
_STATE_TRIGGERS = (
    (CONF_ON_STATE, None),
    (CONF_ON_IDLE, MediaPlayerState.MEDIA_PLAYER_STATE_IDLE),
    (CONF_ON_PLAY, MediaPlayerState.MEDIA_PLAYER_STATE_PLAYING),
    (CONF_ON_PAUSE, MediaPlayerState.MEDIA_PLAYER_STATE_PAUSED),
    (CONF_ON_ANNOUNCEMENT, MediaPlayerState.MEDIA_PLAYER_STATE_ANNOUNCING),
    (CONF_ON_TURN_ON, MediaPlayerState.MEDIA_PLAYER_STATE_ON),
    (CONF_ON_TURN_OFF, MediaPlayerState.MEDIA_PLAYER_STATE_OFF),
)

# State conditions that all share the same schema and codegen handler
_STATE_CONDITIONS = [
    "idle",
    "paused",
    "playing",
    "announcing",
    "on",
    "off",
    "muted",
]

# Special action classes with custom schemas/handlers
PlayMediaAction = media_player_ns.class_(
    "PlayMediaAction", automation.Action, cg.Parented.template(MediaPlayer)
)
EnqueueMediaAction = media_player_ns.class_(
    "EnqueueMediaAction", automation.Action, cg.Parented.template(MediaPlayer)
)
VolumeSetAction = media_player_ns.class_(
    "VolumeSetAction", automation.Action, cg.Parented.template(MediaPlayer)
)


_CALLBACK_AUTOMATIONS = (
    automation.CallbackAutomation(
        CONF_ON_STATE, "add_on_state_callback", forwarder=StateAnyForwarder
    ),
    automation.CallbackAutomation(
        CONF_ON_IDLE,
        "add_on_state_callback",
        forwarder=StateEnterForwarder.template(
            MediaPlayerState.MEDIA_PLAYER_STATE_IDLE
        ),
    ),
    automation.CallbackAutomation(
        CONF_ON_PLAY,
        "add_on_state_callback",
        forwarder=StateEnterForwarder.template(
            MediaPlayerState.MEDIA_PLAYER_STATE_PLAYING
        ),
    ),
    automation.CallbackAutomation(
        CONF_ON_PAUSE,
        "add_on_state_callback",
        forwarder=StateEnterForwarder.template(
            MediaPlayerState.MEDIA_PLAYER_STATE_PAUSED
        ),
    ),
    automation.CallbackAutomation(
        CONF_ON_ANNOUNCEMENT,
        "add_on_state_callback",
        forwarder=StateEnterForwarder.template(
            MediaPlayerState.MEDIA_PLAYER_STATE_ANNOUNCING
        ),
    ),
    automation.CallbackAutomation(
        CONF_ON_TURN_ON,
        "add_on_state_callback",
        forwarder=StateEnterForwarder.template(MediaPlayerState.MEDIA_PLAYER_STATE_ON),
    ),
    automation.CallbackAutomation(
        CONF_ON_TURN_OFF,
        "add_on_state_callback",
        forwarder=StateEnterForwarder.template(MediaPlayerState.MEDIA_PLAYER_STATE_OFF),
    ),
)


@setup_entity("media_player")
async def setup_media_player_core_(var, config):
    await automation.build_callback_automations(var, config, _CALLBACK_AUTOMATIONS)


async def register_media_player(var, config):
    if not CORE.has_id(config[CONF_ID]):
        var = cg.Pvariable(config[CONF_ID], var)
    queue_entity_register("media_player", config)
    CORE.register_platform_component("media_player", var)
    await setup_media_player_core_(var, config)


async def new_media_player(config, *args):
    var = cg.new_Pvariable(config[CONF_ID], *args)
    await register_media_player(var, config)
    return var


_MEDIA_PLAYER_SCHEMA = cv.ENTITY_BASE_SCHEMA.extend(
    {
        cv.Optional(conf_key): automation.validate_automation({})
        for conf_key, _ in _STATE_TRIGGERS
    }
)

_MEDIA_PLAYER_SCHEMA.add_extra(entity_duplicate_validator("media_player"))


def media_player_schema(
    class_: MockObjClass,
    *,
    entity_category: str = cv.UNDEFINED,
    icon: str = cv.UNDEFINED,
) -> cv.Schema:
    schema = {cv.GenerateID(CONF_ID): cv.declare_id(class_)}

    for key, default, validator in [
        (CONF_ENTITY_CATEGORY, entity_category, cv.entity_category),
        (CONF_ICON, icon, cv.icon),
    ]:
        if default is not cv.UNDEFINED:
            schema[cv.Optional(key, default=default)] = validator

    return _MEDIA_PLAYER_SCHEMA.extend(schema)


MEDIA_PLAYER_ACTION_SCHEMA = automation.maybe_simple_id(
    cv.Schema(
        {
            cv.GenerateID(): cv.use_id(MediaPlayer),
            cv.Optional(CONF_ANNOUNCEMENT, default=False): cv.templatable(cv.boolean),
        }
    )
)

MEDIA_PLAYER_CONDITION_SCHEMA = automation.maybe_simple_id(
    {cv.GenerateID(): cv.use_id(MediaPlayer)}
)


_MEDIA_URL_ACTION_SCHEMA = cv.maybe_simple_value(
    {
        cv.GenerateID(): cv.use_id(MediaPlayer),
        cv.Required(CONF_MEDIA_URL): cv.templatable(cv.url),
        cv.Optional(CONF_ANNOUNCEMENT, default=False): cv.templatable(cv.boolean),
    },
    key=CONF_MEDIA_URL,
)


async def _media_action_handler(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])
    media_url = await cg.templatable(config[CONF_MEDIA_URL], args, cg.std_string)
    announcement = await cg.templatable(config[CONF_ANNOUNCEMENT], args, cg.bool_)
    cg.add(var.set_media_url(media_url))
    cg.add(var.set_announcement(announcement))
    return var


automation.register_action(
    "media_player.play_media",
    PlayMediaAction,
    _MEDIA_URL_ACTION_SCHEMA,
    synchronous=True,
)(_media_action_handler)

automation.register_action(
    "media_player.enqueue",
    EnqueueMediaAction,
    _MEDIA_URL_ACTION_SCHEMA,
    synchronous=True,
)(_media_action_handler)


def _snake_to_camel(name):
    return "".join(word.capitalize() for word in name.split("_"))


def _register_command_actions():
    async def handler(config, action_id, template_arg, args):
        var = cg.new_Pvariable(action_id, template_arg)
        await cg.register_parented(var, config[CONF_ID])
        announcement = await cg.templatable(config[CONF_ANNOUNCEMENT], args, cg.bool_)
        cg.add(var.set_announcement(announcement))
        return var

    for action_name in _COMMAND_ACTIONS:
        class_name = f"{_snake_to_camel(action_name)}Action"
        action_class = media_player_ns.class_(
            class_name, automation.Action, cg.Parented.template(MediaPlayer)
        )
        automation.register_action(
            f"media_player.{action_name}",
            action_class,
            MEDIA_PLAYER_ACTION_SCHEMA,
            synchronous=True,
        )(handler)


_register_command_actions()


def _register_state_conditions():
    async def handler(config, action_id, template_arg, args):
        var = cg.new_Pvariable(action_id, template_arg)
        await cg.register_parented(var, config[CONF_ID])
        return var

    for condition_name in _STATE_CONDITIONS:
        class_name = f"Is{_snake_to_camel(condition_name)}Condition"
        condition_class = media_player_ns.class_(class_name, automation.Condition)
        automation.register_condition(
            f"media_player.is_{condition_name}",
            condition_class,
            MEDIA_PLAYER_CONDITION_SCHEMA,
        )(handler)


_register_state_conditions()


@automation.register_action(
    "media_player.volume_set",
    VolumeSetAction,
    cv.maybe_simple_value(
        {
            cv.GenerateID(): cv.use_id(MediaPlayer),
            cv.Required(CONF_VOLUME): cv.templatable(cv.percentage),
        },
        key=CONF_VOLUME,
    ),
    synchronous=True,
)
async def media_player_volume_set_action(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])
    volume = await cg.templatable(config[CONF_VOLUME], args, cg.float_)
    cg.add(var.set_volume(volume))
    return var


@coroutine_with_priority(CoroPriority.CORE)
async def to_code(config):
    cg.add_global(media_player_ns.using)
    cg.add_define("USE_MEDIA_PLAYER")



import esphome.codegen as cg
from esphome.components import binary_sensor
from esphome.components.binary_sensor import BinarySensor
import esphome.config_validation as cv
from esphome.const import CONF_KEY
from esphome.core import Lambda
from esphome.cpp_generator import ExpressionStatement, RawExpression

from .display import CONF_SDL_ID, Sdl

CODEOWNERS = ["@bdm310"]

STATE_ARG = "state"

SDL_KeyCode = cg.global_ns.enum("SDL_KeyCode")

SDL_KEYS = (
    "SDLK_UNKNOWN",
    "SDLK_RETURN",
    "SDLK_ESCAPE",
    "SDLK_BACKSPACE",
    "SDLK_TAB",
    "SDLK_SPACE",
    "SDLK_EXCLAIM",
    "SDLK_QUOTEDBL",
    "SDLK_HASH",
    "SDLK_PERCENT",
    "SDLK_DOLLAR",
    "SDLK_AMPERSAND",
    "SDLK_QUOTE",
    "SDLK_LEFTPAREN",
    "SDLK_RIGHTPAREN",
    "SDLK_ASTERISK",
    "SDLK_PLUS",
    "SDLK_COMMA",
    "SDLK_MINUS",
    "SDLK_PERIOD",
    "SDLK_SLASH",
    "SDLK_0",
    "SDLK_1",
    "SDLK_2",
    "SDLK_3",
    "SDLK_4",
    "SDLK_5",
    "SDLK_6",
    "SDLK_7",
    "SDLK_8",
    "SDLK_9",
    "SDLK_COLON",
    "SDLK_SEMICOLON",
    "SDLK_LESS",
    "SDLK_EQUALS",
    "SDLK_GREATER",
    "SDLK_QUESTION",
    "SDLK_AT",
    "SDLK_LEFTBRACKET",
    "SDLK_BACKSLASH",
    "SDLK_RIGHTBRACKET",
    "SDLK_CARET",
    "SDLK_UNDERSCORE",
    "SDLK_BACKQUOTE",
    "SDLK_a",
    "SDLK_b",
    "SDLK_c",
    "SDLK_d",
    "SDLK_e",
    "SDLK_f",
    "SDLK_g",
    "SDLK_h",
    "SDLK_i",
    "SDLK_j",
    "SDLK_k",
    "SDLK_l",
    "SDLK_m",
    "SDLK_n",
    "SDLK_o",
    "SDLK_p",
    "SDLK_q",
    "SDLK_r",
    "SDLK_s",
    "SDLK_t",
    "SDLK_u",
    "SDLK_v",
    "SDLK_w",
    "SDLK_x",
    "SDLK_y",
    "SDLK_z",
    "SDLK_CAPSLOCK",
    "SDLK_F1",
    "SDLK_F2",
    "SDLK_F3",
    "SDLK_F4",
    "SDLK_F5",
    "SDLK_F6",
    "SDLK_F7",
    "SDLK_F8",
    "SDLK_F9",
    "SDLK_F10",
    "SDLK_F11",
    "SDLK_F12",
    "SDLK_PRINTSCREEN",
    "SDLK_SCROLLLOCK",
    "SDLK_PAUSE",
    "SDLK_INSERT",
    "SDLK_HOME",
    "SDLK_PAGEUP",
    "SDLK_DELETE",
    "SDLK_END",
    "SDLK_PAGEDOWN",
    "SDLK_RIGHT",
    "SDLK_LEFT",
    "SDLK_DOWN",
    "SDLK_UP",
    "SDLK_NUMLOCKCLEAR",
    "SDLK_KP_DIVIDE",
    "SDLK_KP_MULTIPLY",
    "SDLK_KP_MINUS",
    "SDLK_KP_PLUS",
    "SDLK_KP_ENTER",
    "SDLK_KP_1",
    "SDLK_KP_2",
    "SDLK_KP_3",
    "SDLK_KP_4",
    "SDLK_KP_5",
    "SDLK_KP_6",
    "SDLK_KP_7",
    "SDLK_KP_8",
    "SDLK_KP_9",
    "SDLK_KP_0",
    "SDLK_KP_PERIOD",
    "SDLK_APPLICATION",
    "SDLK_POWER",
    "SDLK_KP_EQUALS",
    "SDLK_F13",
    "SDLK_F14",
    "SDLK_F15",
    "SDLK_F16",
    "SDLK_F17",
    "SDLK_F18",
    "SDLK_F19",
    "SDLK_F20",
    "SDLK_F21",
    "SDLK_F22",
    "SDLK_F23",
    "SDLK_F24",
    "SDLK_EXECUTE",
    "SDLK_HELP",
    "SDLK_MENU",
    "SDLK_SELECT",
    "SDLK_STOP",
    "SDLK_AGAIN",
    "SDLK_UNDO",
    "SDLK_CUT",
    "SDLK_COPY",
    "SDLK_PASTE",
    "SDLK_FIND",
    "SDLK_MUTE",
    "SDLK_VOLUMEUP",
    "SDLK_VOLUMEDOWN",
    "SDLK_KP_COMMA",
    "SDLK_KP_EQUALSAS400",
    "SDLK_ALTERASE",
    "SDLK_SYSREQ",
    "SDLK_CANCEL",
    "SDLK_CLEAR",
    "SDLK_PRIOR",
    "SDLK_RETURN2",
    "SDLK_SEPARATOR",
    "SDLK_OUT",
    "SDLK_OPER",
    "SDLK_CLEARAGAIN",
    "SDLK_CRSEL",
    "SDLK_EXSEL",
    "SDLK_KP_00",
    "SDLK_KP_000",
    "SDLK_THOUSANDSSEPARATOR",
    "SDLK_DECIMALSEPARATOR",
    "SDLK_CURRENCYUNIT",
    "SDLK_CURRENCYSUBUNIT",
    "SDLK_KP_LEFTPAREN",
    "SDLK_KP_RIGHTPAREN",
    "SDLK_KP_LEFTBRACE",
    "SDLK_KP_RIGHTBRACE",
    "SDLK_KP_TAB",
    "SDLK_KP_BACKSPACE",
    "SDLK_KP_A",
    "SDLK_KP_B",
    "SDLK_KP_C",
    "SDLK_KP_D",
    "SDLK_KP_E",
    "SDLK_KP_F",
    "SDLK_KP_XOR",
    "SDLK_KP_POWER",
    "SDLK_KP_PERCENT",
    "SDLK_KP_LESS",
    "SDLK_KP_GREATER",
    "SDLK_KP_AMPERSAND",
    "SDLK_KP_DBLAMPERSAND",
    "SDLK_KP_VERTICALBAR",
    "SDLK_KP_DBLVERTICALBAR",
    "SDLK_KP_COLON",
    "SDLK_KP_HASH",
    "SDLK_KP_SPACE",
    "SDLK_KP_AT",
    "SDLK_KP_EXCLAM",
    "SDLK_KP_MEMSTORE",
    "SDLK_KP_MEMRECALL",
    "SDLK_KP_MEMCLEAR",
    "SDLK_KP_MEMADD",
    "SDLK_KP_MEMSUBTRACT",
    "SDLK_KP_MEMMULTIPLY",
    "SDLK_KP_MEMDIVIDE",
    "SDLK_KP_PLUSMINUS",
    "SDLK_KP_CLEAR",
    "SDLK_KP_CLEARENTRY",
    "SDLK_KP_BINARY",
    "SDLK_KP_OCTAL",
    "SDLK_KP_DECIMAL",
    "SDLK_KP_HEXADECIMAL",
    "SDLK_LCTRL",
    "SDLK_LSHIFT",
    "SDLK_LALT",
    "SDLK_LGUI",
    "SDLK_RCTRL",
    "SDLK_RSHIFT",
    "SDLK_RALT",
    "SDLK_RGUI",
    "SDLK_MODE",
    "SDLK_AUDIONEXT",
    "SDLK_AUDIOPREV",
    "SDLK_AUDIOSTOP",
    "SDLK_AUDIOPLAY",
    "SDLK_AUDIOMUTE",
    "SDLK_MEDIASELECT",
    "SDLK_WWW",
    "SDLK_MAIL",
    "SDLK_CALCULATOR",
    "SDLK_COMPUTER",
    "SDLK_AC_SEARCH",
    "SDLK_AC_HOME",
    "SDLK_AC_BACK",
    "SDLK_AC_FORWARD",
    "SDLK_AC_STOP",
    "SDLK_AC_REFRESH",
    "SDLK_AC_BOOKMARKS",
    "SDLK_BRIGHTNESSDOWN",
    "SDLK_BRIGHTNESSUP",
    "SDLK_DISPLAYSWITCH",
    "SDLK_KBDILLUMTOGGLE",
    "SDLK_KBDILLUMDOWN",
    "SDLK_KBDILLUMUP",
    "SDLK_EJECT",
    "SDLK_SLEEP",
    "SDLK_APP1",
    "SDLK_APP2",
    "SDLK_AUDIOREWIND",
    "SDLK_AUDIOFASTFORWARD",
    "SDLK_SOFTLEFT",
    "SDLK_SOFTRIGHT",
    "SDLK_CALL",
    "SDLK_ENDCALL",
)

SDL_KEYMAP = {key: getattr(SDL_KeyCode, key) for key in SDL_KEYS}

CONFIG_SCHEMA = (
    binary_sensor.binary_sensor_schema(BinarySensor)
    .extend(
        {
            cv.Required(CONF_KEY): cv.enum(SDL_KEYMAP),
            cv.GenerateID(CONF_SDL_ID): cv.use_id(Sdl),
        }
    )
    .extend(cv.COMPONENT_SCHEMA)
)


async def to_code(config):
    var = await binary_sensor.new_binary_sensor(config)
    parent = await cg.get_variable(config[CONF_SDL_ID])
    listener = Lambda(
        str(ExpressionStatement(var.publish_state(RawExpression(STATE_ARG))))
    )
    listener = await cg.process_lambda(listener, [(cg.bool_, STATE_ARG)])
    cg.add(parent.add_key_listener(config[CONF_KEY], listener))



import esphome.codegen as cg
from esphome.components import switch
import esphome.config_validation as cv
from esphome.const import (
    CONF_BLUETOOTH,
    CONF_ID,
    DEVICE_CLASS_SWITCH,
    ENTITY_CATEGORY_CONFIG,
    ICON_BLUETOOTH,
    ICON_PULSE,
)

from .. import CONF_LD2450_ID, LD2450Component, ld2450_ns

BluetoothSwitch = ld2450_ns.class_("BluetoothSwitch", switch.Switch)
MultiTargetSwitch = ld2450_ns.class_("MultiTargetSwitch", switch.Switch)

CONF_MULTI_TARGET = "multi_target"

CONFIG_SCHEMA = {
    cv.GenerateID(CONF_ID): cv.declare_id(cg.EntityBase),
    cv.GenerateID(CONF_LD2450_ID): cv.use_id(LD2450Component),
    cv.Optional(CONF_BLUETOOTH): switch.switch_schema(
        BluetoothSwitch,
        device_class=DEVICE_CLASS_SWITCH,
        entity_category=ENTITY_CATEGORY_CONFIG,
        icon=ICON_BLUETOOTH,
    ),
    cv.Optional(CONF_MULTI_TARGET): switch.switch_schema(
        MultiTargetSwitch,
        device_class=DEVICE_CLASS_SWITCH,
        entity_category=ENTITY_CATEGORY_CONFIG,
        icon=ICON_PULSE,
    ),
}


async def to_code(config):
    ld2450_component = await cg.get_variable(config[CONF_LD2450_ID])
    if bluetooth_config := config.get(CONF_BLUETOOTH):
        s = await switch.new_switch(bluetooth_config)
        await cg.register_parented(s, config[CONF_LD2450_ID])
        cg.add(ld2450_component.set_bluetooth_switch(s))
    if multi_target_config := config.get(CONF_MULTI_TARGET):
        s = await switch.new_switch(multi_target_config)
        await cg.register_parented(s, config[CONF_LD2450_ID])
        cg.add(ld2450_component.set_multi_target_switch(s))



import esphome.codegen as cg
from esphome.components import media_player
from esphome.components.const import CONF_VOLUME_INCREMENT
import esphome.config_validation as cv
from esphome.const import CONF_ID
from esphome.types import ConfigType

from .. import CONF_SENDSPIN_ID, SendspinHub, request_controller_support, sendspin_ns

CODEOWNERS = ["@kahrendt"]
DEPENDENCIES = ["sendspin"]

SendspinMediaPlayer = sendspin_ns.class_(
    "SendspinMediaPlayer",
    media_player.MediaPlayer,
    cg.Component,
)


def _request_roles(config: ConfigType) -> ConfigType:
    """Request the necessary Sendspin roles for the media player."""
    request_controller_support()

    return config


CONFIG_SCHEMA = cv.All(
    media_player.media_player_schema(SendspinMediaPlayer).extend(
        {
            cv.GenerateID(CONF_SENDSPIN_ID): cv.use_id(SendspinHub),
            cv.Optional(CONF_VOLUME_INCREMENT, default=0.05): cv.percentage,
        }
    ),
    cv.only_on_esp32,
    _request_roles,
)


async def to_code(config: ConfigType) -> None:
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)
    await cg.register_parented(var, config[CONF_SENDSPIN_ID])
    await media_player.register_media_player(var, config)

    cg.add(var.set_volume_increment(config[CONF_VOLUME_INCREMENT]))



import esphome.codegen as cg
from esphome.components import sensor
import esphome.config_validation as cv
from esphome.const import (
    CONF_ID,
    CONF_MOVING_DISTANCE,
    DEVICE_CLASS_DISTANCE,
    STATE_CLASS_MEASUREMENT,
    UNIT_CENTIMETER,
)

from .. import CONF_LD2420_ID, LD2420Component, ld2420_ns

LD2420Sensor = ld2420_ns.class_("LD2420Sensor", sensor.Sensor, cg.Component)

CONF_GATE_ENERGY = "gate_energy"

CONFIG_SCHEMA = cv.All(
    cv.COMPONENT_SCHEMA.extend(
        {
            cv.GenerateID(): cv.declare_id(LD2420Sensor),
            cv.GenerateID(CONF_LD2420_ID): cv.use_id(LD2420Component),
            cv.Optional(CONF_MOVING_DISTANCE): sensor.sensor_schema(
                device_class=DEVICE_CLASS_DISTANCE,
                unit_of_measurement=UNIT_CENTIMETER,
                state_class=STATE_CLASS_MEASUREMENT,
            ),
        }
    ),
)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)
    if CONF_MOVING_DISTANCE in config:
        sens = await sensor.new_sensor(config[CONF_MOVING_DISTANCE])
        cg.add(var.set_distance_sensor(sens))
    if CONF_GATE_ENERGY in config:
        sens = await sensor.new_sensor(config[CONF_GATE_ENERGY])
        cg.add(var.set_energy_sensor(sens))
    ld2420 = await cg.get_variable(config[CONF_LD2420_ID])
    cg.add(ld2420.register_listener(var))



from esphome import automation, pins
from esphome.automation import maybe_simple_id
import esphome.codegen as cg
from esphome.components import nfc
import esphome.config_validation as cv
from esphome.const import (
    CONF_ID,
    CONF_IRQ_PIN,
    CONF_MESSAGE,
    CONF_ON_FINISHED_WRITE,
    CONF_ON_TAG,
    CONF_ON_TAG_REMOVED,
    CONF_TRIGGER_ID,
)

AUTO_LOAD = ["binary_sensor", "nfc"]
CODEOWNERS = ["@kbx81", "@jesserockz"]

CONF_DWL_REQ_PIN = "dwl_req_pin"
CONF_EMULATION_MESSAGE = "emulation_message"
CONF_EMULATION_OFF = "emulation_off"
CONF_EMULATION_ON = "emulation_on"
CONF_INCLUDE_ANDROID_APP_RECORD = "include_android_app_record"
CONF_ON_EMULATED_TAG_SCAN = "on_emulated_tag_scan"
CONF_PN7160_ID = "pn7160_id"
CONF_POLLING_OFF = "polling_off"
CONF_POLLING_ON = "polling_on"
CONF_SET_CLEAN_MODE = "set_clean_mode"
CONF_SET_EMULATION_MESSAGE = "set_emulation_message"
CONF_SET_FORMAT_MODE = "set_format_mode"
CONF_SET_READ_MODE = "set_read_mode"
CONF_SET_WRITE_MESSAGE = "set_write_message"
CONF_SET_WRITE_MODE = "set_write_mode"
CONF_TAG_TTL = "tag_ttl"
CONF_VEN_PIN = "ven_pin"
CONF_WKUP_REQ_PIN = "wkup_req_pin"

pn7160_ns = cg.esphome_ns.namespace("pn7160")
PN7160 = pn7160_ns.class_("PN7160", nfc.Nfcc, cg.Component)

EmulationOffAction = pn7160_ns.class_("EmulationOffAction", automation.Action)
EmulationOnAction = pn7160_ns.class_("EmulationOnAction", automation.Action)
PollingOffAction = pn7160_ns.class_("PollingOffAction", automation.Action)
PollingOnAction = pn7160_ns.class_("PollingOnAction", automation.Action)
SetCleanModeAction = pn7160_ns.class_("SetCleanModeAction", automation.Action)
SetEmulationMessageAction = pn7160_ns.class_(
    "SetEmulationMessageAction", automation.Action
)
SetFormatModeAction = pn7160_ns.class_("SetFormatModeAction", automation.Action)
SetReadModeAction = pn7160_ns.class_("SetReadModeAction", automation.Action)
SetWriteMessageAction = pn7160_ns.class_("SetWriteMessageAction", automation.Action)
SetWriteModeAction = pn7160_ns.class_("SetWriteModeAction", automation.Action)


PN7160IsWritingCondition = pn7160_ns.class_(
    "PN7160IsWritingCondition", automation.Condition
)


IsWritingCondition = nfc.nfc_ns.class_("IsWritingCondition", automation.Condition)


SIMPLE_ACTION_SCHEMA = maybe_simple_id(
    {
        cv.Required(CONF_ID): cv.use_id(PN7160),
    }
)

SET_MESSAGE_ACTION_SCHEMA = cv.Schema(
    {
        cv.GenerateID(): cv.use_id(PN7160),
        cv.Required(CONF_MESSAGE): cv.templatable(cv.string),
        cv.Optional(CONF_INCLUDE_ANDROID_APP_RECORD, default=True): cv.boolean,
    }
)

PN7160_SCHEMA = cv.Schema(
    {
        cv.GenerateID(): cv.declare_id(PN7160),
        cv.Optional(CONF_ON_EMULATED_TAG_SCAN): automation.validate_automation({}),
        cv.Optional(CONF_ON_FINISHED_WRITE): automation.validate_automation({}),
        cv.Optional(CONF_ON_TAG): automation.validate_automation(
            {
                cv.GenerateID(CONF_TRIGGER_ID): cv.declare_id(nfc.NfcOnTagTrigger),
            }
        ),
        cv.Optional(CONF_ON_TAG_REMOVED): automation.validate_automation(
            {
                cv.GenerateID(CONF_TRIGGER_ID): cv.declare_id(nfc.NfcOnTagTrigger),
            }
        ),
        cv.Optional(CONF_DWL_REQ_PIN): pins.gpio_output_pin_schema,
        cv.Required(CONF_IRQ_PIN): pins.gpio_input_pin_schema,
        cv.Required(CONF_VEN_PIN): pins.gpio_output_pin_schema,
        cv.Optional(CONF_WKUP_REQ_PIN): pins.gpio_output_pin_schema,
        cv.Optional(CONF_EMULATION_MESSAGE): cv.string,
        cv.Optional(CONF_TAG_TTL): cv.positive_time_period_milliseconds,
    }
).extend(cv.COMPONENT_SCHEMA)


@automation.register_action(
    "tag.set_emulation_message",
    SetEmulationMessageAction,
    SET_MESSAGE_ACTION_SCHEMA,
    synchronous=True,
)
@automation.register_action(
    "tag.set_write_message",
    SetWriteMessageAction,
    SET_MESSAGE_ACTION_SCHEMA,
    synchronous=True,
)
async def pn7160_set_message_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])
    template_ = await cg.templatable(config[CONF_MESSAGE], args, cg.std_string)
    cg.add(var.set_message(template_))
    template_ = await cg.templatable(
        config[CONF_INCLUDE_ANDROID_APP_RECORD], args, cg.bool_
    )
    cg.add(var.set_include_android_app_record(template_))
    return var


@automation.register_action(
    "tag.emulation_off",
    EmulationOffAction,
    SIMPLE_ACTION_SCHEMA,
    synchronous=True,
)
@automation.register_action(
    "tag.emulation_on", EmulationOnAction, SIMPLE_ACTION_SCHEMA, synchronous=True
)
@automation.register_action(
    "tag.polling_off", PollingOffAction, SIMPLE_ACTION_SCHEMA, synchronous=True
)
@automation.register_action(
    "tag.polling_on", PollingOnAction, SIMPLE_ACTION_SCHEMA, synchronous=True
)
@automation.register_action(
    "tag.set_clean_mode",
    SetCleanModeAction,
    SIMPLE_ACTION_SCHEMA,
    synchronous=True,
)
@automation.register_action(
    "tag.set_format_mode",
    SetFormatModeAction,
    SIMPLE_ACTION_SCHEMA,
    synchronous=True,
)
@automation.register_action(
    "tag.set_read_mode",
    SetReadModeAction,
    SIMPLE_ACTION_SCHEMA,
    synchronous=True,
)
@automation.register_action(
    "tag.set_write_mode",
    SetWriteModeAction,
    SIMPLE_ACTION_SCHEMA,
    synchronous=True,
)
async def pn7160_simple_action_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])
    return var


_CALLBACK_AUTOMATIONS = (
    automation.CallbackAutomation(
        CONF_ON_EMULATED_TAG_SCAN, "add_on_emulated_tag_scan_callback"
    ),
    automation.CallbackAutomation(
        CONF_ON_FINISHED_WRITE, "add_on_finished_write_callback"
    ),
)


async def setup_pn7160(var, config):
    await cg.register_component(var, config)

    if dwl_req_pin_config := config.get(CONF_DWL_REQ_PIN):
        pin = await cg.gpio_pin_expression(dwl_req_pin_config)
        cg.add(var.set_dwl_req_pin(pin))

    pin = await cg.gpio_pin_expression(config[CONF_IRQ_PIN])
    cg.add(var.set_irq_pin(pin))

    pin = await cg.gpio_pin_expression(config[CONF_VEN_PIN])
    cg.add(var.set_ven_pin(pin))

    if wakeup_req_pin_config := config.get(CONF_WKUP_REQ_PIN):
        pin = await cg.gpio_pin_expression(wakeup_req_pin_config)
        cg.add(var.set_wkup_req_pin(pin))

    if emulation_message_config := config.get(CONF_EMULATION_MESSAGE):
        cg.add(var.set_tag_emulation_message(emulation_message_config))
        cg.add(var.set_tag_emulation_on())

    if CONF_TAG_TTL in config:
        cg.add(var.set_tag_ttl(config[CONF_TAG_TTL]))

    for conf in config.get(CONF_ON_TAG, []):
        trigger = cg.new_Pvariable(conf[CONF_TRIGGER_ID])
        cg.add(var.register_ontag_trigger(trigger))
        await automation.build_automation(
            trigger, [(cg.std_string, "x"), (nfc.NfcTag, "tag")], conf
        )

    for conf in config.get(CONF_ON_TAG_REMOVED, []):
        trigger = cg.new_Pvariable(conf[CONF_TRIGGER_ID])
        cg.add(var.register_ontagremoved_trigger(trigger))
        await automation.build_automation(
            trigger, [(cg.std_string, "x"), (nfc.NfcTag, "tag")], conf
        )

    await automation.build_callback_automations(var, config, _CALLBACK_AUTOMATIONS)


@automation.register_condition(
    "pn7160.is_writing",
    PN7160IsWritingCondition,
    cv.Schema(
        {
            cv.GenerateID(): cv.use_id(PN7160),
        }
    ),
)
async def pn7160_is_writing_to_code(config, condition_id, template_arg, args):
    var = cg.new_Pvariable(condition_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])
    return var



"""
ESPHome configuration for the IT8951 e-paper controller.
"""

from esphome import automation, core, pins
import esphome.codegen as cg
from esphome.components import display, spi
from esphome.components.display import CONF_SHOW_TEST_CARD, validate_rotation
import esphome.config_validation as cv
from esphome.config_validation import update_interval
from esphome.const import (
    CONF_BUSY_PIN,
    CONF_CS_PIN,
    CONF_DATA_RATE,
    CONF_DIMENSIONS,
    CONF_ENABLE_PIN,
    CONF_FULL_UPDATE_EVERY,
    CONF_HEIGHT,
    CONF_ID,
    CONF_INVERT_COLORS,
    CONF_LAMBDA,
    CONF_MIRROR_X,
    CONF_MIRROR_Y,
    CONF_MODE,
    CONF_MODEL,
    CONF_PAGES,
    CONF_RESET_DURATION,
    CONF_RESET_PIN,
    CONF_ROTATION,
    CONF_SLEEP_WHEN_DONE,
    CONF_SWAP_XY,
    CONF_TRANSFORM,
    CONF_UPDATE_INTERVAL,
    CONF_WIDTH,
)
from esphome.cpp_generator import RawExpression
from esphome.final_validate import full_config

AUTO_LOAD = ["split_buffer"]
DEPENDENCIES = ["spi"]

CONF_VCOM = "vcom"
CONF_VCOM_REGISTER = "vcom_register"
CONF_FORCE_TEMPERATURE = "force_temperature"
CONF_GRAYSCALE = "grayscale"
CONF_DITHERING = "dithering"
CONF_UPDATE_MODE = "update_mode"
CONF_USE_LEGACY_DPY_AREA = "use_legacy_dpy_area"

# VCOM SET sub-command selectors. The IT8951 firmware accepts different
# values across panels; most respond to 0x0001, but a few — e.g. the Seeed
# reTerminal E1003 — only respond to 0x0002 and silently drop 0x0001.
VCOM_REGISTER_DEFAULT = 0x0001
VCOM_REGISTER_ALT = 0x0002
VCOM_REGISTER_OPTIONS = (VCOM_REGISTER_DEFAULT, VCOM_REGISTER_ALT)

it8951_ns = cg.esphome_ns.namespace("it8951")
IT8951Display = it8951_ns.class_("IT8951Display", display.Display, spi.SPIDevice)
IT8951UpdateAction = it8951_ns.class_("IT8951UpdateAction", automation.Action)

# Hardware waveform modes exposed to YAML. Strings are mapped to the C++
# UpdateMode enum so the runtime can store the mode as a uint16_t rather
# than a std::string (avoiding a heap-resident member; see ESPHome
# CLAUDE.md "STL Container Guidelines"). "fast" and "full" are
# convenience aliases for DU and GC16 respectively.
UpdateMode = it8951_ns.enum("UpdateMode")
UPDATE_MODE_OPTIONS = {
    "INIT": UpdateMode.UPDATE_MODE_INIT,
    "DU": UpdateMode.UPDATE_MODE_DU,
    "GC16": UpdateMode.UPDATE_MODE_GC16,
    "GL16": UpdateMode.UPDATE_MODE_GL16,
    "GLR16": UpdateMode.UPDATE_MODE_GLR16,
    "GLD16": UpdateMode.UPDATE_MODE_GLD16,
    "DU4": UpdateMode.UPDATE_MODE_DU4,
    "A2": UpdateMode.UPDATE_MODE_A2,
    "FAST": UpdateMode.UPDATE_MODE_DU,
    "FULL": UpdateMode.UPDATE_MODE_GC16,
}
# Maps the YAML mode string directly to the C++ UpdateMode enum value, so the
# config option and the it8951.update action share one validator.
update_mode = cv.enum(UPDATE_MODE_OPTIONS, upper=True)

# Transform flag values mirror the C++ TRANSFORM_* constants.
_TRANSFORM_NONE = 0
_TRANSFORM_MIRROR_X = 1
_TRANSFORM_MIRROR_Y = 2
_TRANSFORM_SWAP_XY = 4
_TRANSFORM_FLAGS = {
    CONF_MIRROR_X: _TRANSFORM_MIRROR_X,
    CONF_MIRROR_Y: _TRANSFORM_MIRROR_Y,
    CONF_SWAP_XY: _TRANSFORM_SWAP_XY,
}


class IT8951Model:
    """A specific board / panel preset for the IT8951 controller."""

    models: dict[str, "IT8951Model"] = {}

    def __init__(self, name: str, **defaults):
        name = name.upper()
        self.name = name
        self.defaults = defaults
        IT8951Model.models[name] = self

    def get_default(self, key, fallback=None):
        return self.defaults.get(key, fallback)

    def get_dimensions(self, config) -> tuple[int, int]:
        # If dimensions are in config, use them; otherwise fall back to model defaults.
        if CONF_DIMENSIONS in config:
            dimensions = config[CONF_DIMENSIONS]
            if isinstance(dimensions, dict):
                return dimensions[CONF_WIDTH], dimensions[CONF_HEIGHT]
            return tuple(dimensions)
        # Model must have defaults if dimensions not in config.
        return self.get_default(CONF_WIDTH), self.get_default(CONF_HEIGHT)


# --- Model presets ----------------------------------------------------------
# The generic model leaves dimensions and pin choices up to the user.
IT8951Model("it8951", vcom=2300, sleep_when_done=True, data_rate=12_000_000)

IT8951Model(
    "m5stack-m5paper",
    width=960,
    height=540,
    busy_pin=27,
    reset_pin=23,
    cs_pin=15,
    vcom=2300,
    sleep_when_done=True,
    data_rate=20_000_000,
)

IT8951Model(
    "seeed-reterminal-e1003",
    width=1872,
    height=1404,
    busy_pin=13,
    reset_pin=12,
    cs_pin=10,
    # Board power-enable rails: 1.8V logic supply (GPIO21) and the EPD supply
    # (GPIO11). Driven high during setup so no separate power_supply is needed.
    enable_pin=[21, 11],
    vcom=1400,
    # reTerminal E1003 panel firmware only accepts the 0x0002 VCOM SET
    # selector; using the default 0x0001 leaves VCOM unchanged and breaks
    # grayscale waveforms (GC16/GL16) — INIT still works because it does
    # not depend on VCOM accuracy.
    vcom_register=VCOM_REGISTER_ALT,
    # The reTerminal E1003 ships with on-die temperature sensing disabled,
    # so the host must declare an operating temperature; otherwise the
    # waveform LUT defaults to a value that produces no visible change
    # for grayscale modes.
    force_temperature=25,
    sleep_when_done=False,
    data_rate=20_000_000,
    mirror_x=True,
)

IT8951Model(
    "seeed-ee03",
    width=1872,
    height=1404,
    busy_pin=4,
    reset_pin=38,
    cs_pin=44,
    vcom=1400,
    sleep_when_done=False,
    data_rate=4_000_000,
)

# ---------------------------------------------------------------------------

DIMENSION_SCHEMA = cv.Schema(
    {
        cv.Required(CONF_WIDTH): cv.int_,
        cv.Required(CONF_HEIGHT): cv.int_,
    }
)


def _model_pin_option(model, key, schema):
    default = model.get_default(key)
    if default is None:
        return cv.Required(key), schema
    return cv.Optional(key, default=default), schema


def _model_schema(config):
    model = IT8951Model.models[config[CONF_MODEL]]
    has_default_dimensions = (
        model.get_default(CONF_WIDTH) is not None
        and model.get_default(CONF_HEIGHT) is not None
    )
    dimensions_key = (
        cv.Optional(
            CONF_DIMENSIONS,
            default={
                CONF_WIDTH: model.get_default(CONF_WIDTH),
                CONF_HEIGHT: model.get_default(CONF_HEIGHT),
            },
        )
        if has_default_dimensions
        else cv.Required(CONF_DIMENSIONS)
    )

    schema = display.FULL_DISPLAY_SCHEMA.extend(
        spi.spi_device_schema(
            cs_pin_required=False,
            default_mode="MODE0",
            default_data_rate=model.get_default(CONF_DATA_RATE, 10_000_000),
        )
    ).extend(
        {
            cv.GenerateID(): cv.declare_id(IT8951Display),
            cv.Required(CONF_MODEL): cv.one_of(model.name, upper=True, space="-"),
            cv.Optional(CONF_ROTATION, default=0): validate_rotation,
            cv.Optional(CONF_UPDATE_INTERVAL, default=cv.UNDEFINED): update_interval,
            cv.Optional(CONF_FULL_UPDATE_EVERY, default=30): cv.int_range(1, 255),
            cv.Optional(CONF_TRANSFORM): cv.Schema(
                {
                    cv.Required(CONF_MIRROR_X): cv.boolean,
                    cv.Required(CONF_MIRROR_Y): cv.boolean,
                    cv.Optional(CONF_SWAP_XY, default=False): cv.boolean,
                }
            ),
            cv.Optional(
                CONF_INVERT_COLORS, default=model.get_default(CONF_INVERT_COLORS, False)
            ): cv.boolean,
            cv.Optional(
                CONF_SLEEP_WHEN_DONE,
                default=model.get_default(CONF_SLEEP_WHEN_DONE, False),
            ): cv.boolean,
            # Pixel format: true = 4bpp grayscale, false = packed 1bpp
            # monochrome. Monochrome halves the framebuffer and enables fast DU
            # partial refreshes; grayscale gives 16 levels but always uses GC16.
            cv.Optional(
                CONF_GRAYSCALE, default=model.get_default(CONF_GRAYSCALE, True)
            ): cv.boolean,
            # Monochrome only: ordered-dither pale colours so they render as
            # visible stipple. Disable for a crisp hard black/white threshold
            # (better for purely black/white text). No effect in grayscale mode.
            cv.Optional(
                CONF_DITHERING, default=model.get_default(CONF_DITHERING, True)
            ): cv.boolean,
            cv.Optional(
                CONF_VCOM, default=model.get_default(CONF_VCOM, 2300)
            ): cv.int_range(0, 5000),
            cv.Optional(
                CONF_VCOM_REGISTER,
                default=model.get_default(CONF_VCOM_REGISTER, VCOM_REGISTER_DEFAULT),
            ): cv.one_of(*VCOM_REGISTER_OPTIONS, int=True),
            **(
                {
                    cv.Optional(
                        CONF_FORCE_TEMPERATURE,
                        default=model.get_default(CONF_FORCE_TEMPERATURE),
                    ): cv.int_range(min=-40, max=85)
                }
                if model.get_default(CONF_FORCE_TEMPERATURE) is not None
                else {}
            ),
            cv.Optional(
                CONF_USE_LEGACY_DPY_AREA,
                default=model.get_default(CONF_USE_LEGACY_DPY_AREA, False),
            ): cv.boolean,
            cv.Optional(CONF_UPDATE_MODE): update_mode,
            # One or more GPIOs driven high during setup to power on the panel
            # (e.g. board power-enable rails), before reset and init.
            cv.Optional(
                CONF_ENABLE_PIN, default=model.get_default(CONF_ENABLE_PIN, [])
            ): cv.ensure_list(pins.gpio_output_pin_schema),
            cv.Optional(CONF_RESET_DURATION): cv.All(
                cv.positive_time_period_milliseconds,
                cv.Range(max=core.TimePeriod(milliseconds=500)),
            ),
            dimensions_key: DIMENSION_SCHEMA,
        }
    )

    # Pin options: required if the model doesn't supply a default.
    pin_specs = (
        (CONF_BUSY_PIN, pins.gpio_input_pin_schema),
        (CONF_RESET_PIN, pins.gpio_output_pin_schema),
        (CONF_CS_PIN, pins.gpio_output_pin_schema),
    )
    pin_extra = {}
    for key, schema_value in pin_specs:
        opt, sv = _model_pin_option(model, key, schema_value)
        pin_extra[opt] = sv
    return schema.extend(pin_extra)


def _customise_schema(config):
    config = cv.Schema(
        {
            cv.Required(CONF_MODEL): cv.one_of(
                *IT8951Model.models, upper=True, space="-"
            )
        },
        extra=cv.ALLOW_EXTRA,
    )(config)

    model_config = _model_schema(config)(config)

    model = IT8951Model.models[config[CONF_MODEL].upper()]
    width, height = model.get_dimensions(model_config)

    display.add_metadata(
        model_config[CONF_ID],
        width,
        height,
        # Rotation is applied per-pixel in draw_pixel_at at no extra cost, so we
        # advertise hardware rotation: LVGL routes its rotation to the driver via
        # set_rotation rather than rotating the framebuffer in software.
        has_hardware_rotation=True,
        has_writer=any(
            model_config.get(key)
            for key in (CONF_LAMBDA, CONF_PAGES, CONF_SHOW_TEST_CARD)
        ),
        # Report the configured rotation so LVGL can detect (and reject) a
        # rotation set in the display config instead of the LVGL config.
        rotation=model_config.get(CONF_ROTATION, 0),
        # The IT8951 snaps partial display refreshes to a 32-pixel X boundary
        # (see prepare_update_region_), so have LVGL round its redraw areas to
        # 32px too — this keeps flush rectangles aligned with what the panel
        # actually refreshes and avoids redundant re-rounding/over-draw.
        draw_rounding=32,
    )

    return model_config


CONFIG_SCHEMA = _customise_schema


def _final_validate(config):
    # IT8951 reads from SPI (DevInfo, VCOM, register reads) so MISO is required.
    spi.final_validate_device_schema("it8951", require_miso=True, require_mosi=True)(
        config
    )

    global_config = full_config.get()
    from esphome.components.lvgl import DOMAIN as LVGL_DOMAIN

    if CONF_LAMBDA not in config and CONF_PAGES not in config:
        if LVGL_DOMAIN in global_config:
            if CONF_UPDATE_INTERVAL not in config:
                config[CONF_UPDATE_INTERVAL] = update_interval("never")
        else:
            config[CONF_SHOW_TEST_CARD] = True
    return config


FINAL_VALIDATE_SCHEMA = _final_validate


async def to_code(config):
    model = IT8951Model.models[config[CONF_MODEL]]
    width, height = model.get_dimensions(config)

    var = cg.new_Pvariable(config[CONF_ID], model.name, width, height)
    await display.register_display(var, config)
    await spi.register_spi_device(var, config, write_only=False)

    if lambda_config := config.get(CONF_LAMBDA):
        lambda_ = await cg.process_lambda(
            lambda_config, [(display.DisplayRef, "it")], return_type=cg.void
        )
        cg.add(var.set_writer(lambda_))
    if reset_pin := config.get(CONF_RESET_PIN):
        cg.add(var.set_reset_pin(await cg.gpio_pin_expression(reset_pin)))
    if busy_pin := config.get(CONF_BUSY_PIN):
        cg.add(var.set_busy_pin(await cg.gpio_pin_expression(busy_pin)))
    if enable_pins := config.get(CONF_ENABLE_PIN):
        cg.add(
            var.set_enable_pins(
                [await cg.gpio_pin_expression(pin) for pin in enable_pins]
            )
        )
    cg.add(var.set_full_update_every(config[CONF_FULL_UPDATE_EVERY]))
    if (reset_duration := config.get(CONF_RESET_DURATION)) is not None:
        cg.add(var.set_reset_duration(reset_duration))
    if config.get(CONF_INVERT_COLORS):
        cg.add(var.set_invert_colors(True))
    if config.get(CONF_SLEEP_WHEN_DONE):
        cg.add(var.set_sleep_when_done(True))
    cg.add(var.set_vcom(config[CONF_VCOM]))
    cg.add(var.set_vcom_register(config[CONF_VCOM_REGISTER]))
    if CONF_FORCE_TEMPERATURE in config:
        cg.add(var.set_force_temperature(config[CONF_FORCE_TEMPERATURE]))
    if config.get(CONF_USE_LEGACY_DPY_AREA):
        cg.add(var.set_use_legacy_dpy_area(True))
    cg.add(var.set_grayscale(config[CONF_GRAYSCALE]))
    cg.add(var.set_dithering(config[CONF_DITHERING]))
    if (mode := config.get(CONF_UPDATE_MODE)) is not None:
        cg.add(var.set_update_mode(mode))

    transform = config.get(
        CONF_TRANSFORM,
        {
            CONF_MIRROR_X: model.get_default(CONF_MIRROR_X),
            CONF_MIRROR_Y: model.get_default(CONF_MIRROR_Y),
        },
    )

    transform_value = sum(
        flag for key, flag in _TRANSFORM_FLAGS.items() if transform.get(key)
    )
    if transform_value:
        cg.add(var.set_transform(RawExpression(str(transform_value))))


@automation.register_action(
    "it8951.update",
    IT8951UpdateAction,
    automation.maybe_simple_id(
        {
            cv.Required(CONF_ID): cv.use_id(IT8951Display),
            cv.Optional(CONF_MODE): cv.templatable(update_mode),
        }
    ),
    synchronous=True,
)
async def it8951_update_action_to_code(config, action_id, template_arg, args):
    display_var = await cg.get_variable(config[CONF_ID])
    var = cg.new_Pvariable(action_id, template_arg, display_var)
    if mode := config.get(CONF_MODE):
        mode = await cg.templatable(mode, args, UpdateMode)
        cg.add(var.set_mode(mode))
    return var



from dataclasses import dataclass

from esphome import automation
from esphome.automation import Trigger
import esphome.codegen as cg
from esphome.components import key_provider
import esphome.config_validation as cv
from esphome.const import (
    CONF_ENABLE_ON_BOOT,
    CONF_ID,
    CONF_MAX_LENGTH,
    CONF_MIN_LENGTH,
    CONF_ON_TIMEOUT,
    CONF_SOURCE_ID,
    CONF_TIMEOUT,
    CONF_TRIGGER_ID,
)
from esphome.cpp_generator import MockObj, literal
from esphome.types import TemplateArgsType

CODEOWNERS = ["@ssieb"]

MULTI_CONF = True

AUTO_LOAD = ["key_provider"]

CONF_START_KEYS = "start_keys"
CONF_END_KEYS = "end_keys"
CONF_END_KEY_REQUIRED = "end_key_required"
CONF_BACK_KEYS = "back_keys"
CONF_CLEAR_KEYS = "clear_keys"
CONF_ALLOWED_KEYS = "allowed_keys"
CONF_ON_PROGRESS = "on_progress"
CONF_ON_RESULT = "on_result"

key_collector_ns = cg.esphome_ns.namespace("key_collector")
KeyCollector = key_collector_ns.class_("KeyCollector", cg.Component)
EnableAction = key_collector_ns.class_("EnableAction", automation.Action)
DisableAction = key_collector_ns.class_("DisableAction", automation.Action)

X_TYPE = cg.std_string_ref.operator("const")


@dataclass
class Argument:
    type: MockObj
    name: str


TRIGGER_TYPES = {
    CONF_ON_PROGRESS: [Argument(X_TYPE, "x"), Argument(cg.uint8, "start")],
    CONF_ON_RESULT: [
        Argument(X_TYPE, "x"),
        Argument(cg.uint8, "start"),
        Argument(cg.uint8, "end"),
    ],
    CONF_ON_TIMEOUT: [Argument(X_TYPE, "x"), Argument(cg.uint8, "start")],
}

CONFIG_SCHEMA = cv.All(
    cv.COMPONENT_SCHEMA.extend(
        {
            cv.GenerateID(): cv.declare_id(KeyCollector),
            cv.Optional(CONF_SOURCE_ID): cv.ensure_list(
                cv.use_id(key_provider.KeyProvider)
            ),
            cv.Optional(CONF_MIN_LENGTH): cv.uint16_t,
            cv.Optional(CONF_MAX_LENGTH): cv.uint16_t,
            cv.Optional(CONF_START_KEYS): cv.string,
            cv.Optional(CONF_END_KEYS): cv.string,
            cv.Optional(CONF_END_KEY_REQUIRED): cv.boolean,
            cv.Optional(CONF_BACK_KEYS): cv.string,
            cv.Optional(CONF_CLEAR_KEYS): cv.string,
            cv.Optional(CONF_ALLOWED_KEYS): cv.string,
            **{
                cv.Optional(trigger_type): automation.validate_automation(
                    {
                        cv.GenerateID(CONF_TRIGGER_ID): cv.declare_id(
                            Trigger.template(*[arg.type for arg in args])
                        ),
                    }
                )
                for trigger_type, args in TRIGGER_TYPES.items()
            },
            cv.Optional(CONF_TIMEOUT): cv.positive_time_period_milliseconds,
            cv.Optional(CONF_ENABLE_ON_BOOT, default=True): cv.boolean,
        }
    ),
    cv.has_at_least_one_key(CONF_END_KEYS, CONF_MAX_LENGTH),
)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)
    for source_conf in config.get(CONF_SOURCE_ID, ()):
        source = await cg.get_variable(source_conf)
        cg.add(var.add_provider(source))
    if CONF_MIN_LENGTH in config:
        cg.add(var.set_min_length(config[CONF_MIN_LENGTH]))
    if CONF_MAX_LENGTH in config:
        cg.add(var.set_max_length(config[CONF_MAX_LENGTH]))
    if CONF_START_KEYS in config:
        cg.add(var.set_start_keys(config[CONF_START_KEYS]))
    if CONF_END_KEYS in config:
        cg.add(var.set_end_keys(config[CONF_END_KEYS]))
    if CONF_END_KEY_REQUIRED in config:
        cg.add(var.set_end_key_required(config[CONF_END_KEY_REQUIRED]))
    if CONF_BACK_KEYS in config:
        cg.add(var.set_back_keys(config[CONF_BACK_KEYS]))
    if CONF_CLEAR_KEYS in config:
        cg.add(var.set_clear_keys(config[CONF_CLEAR_KEYS]))
    if CONF_ALLOWED_KEYS in config:
        cg.add(var.set_allowed_keys(config[CONF_ALLOWED_KEYS]))

    for trigger_name, args in TRIGGER_TYPES.items():
        arglist: TemplateArgsType = [(arg.type, arg.name) for arg in args]
        for conf in config.get(trigger_name, ()):
            trigger = cg.new_Pvariable(conf[CONF_TRIGGER_ID])
            add_trig = getattr(
                var,
                f"add_on_{trigger_name.rsplit('_', maxsplit=1)[-1].lower()}_callback",
            )
            await automation.build_automation(
                trigger,
                arglist,
                conf,
            )
            lamb = trigger.trigger(*[literal(arg.name) for arg in args])
            cg.add(add_trig(await cg.process_lambda(lamb, arglist)))

    if timeout := config.get(CONF_TIMEOUT):
        cg.add(var.set_timeout(timeout))
    cg.add(var.set_enabled(config[CONF_ENABLE_ON_BOOT]))


@automation.register_action(
    "key_collector.enable",
    EnableAction,
    automation.maybe_simple_id(
        {
            cv.GenerateID(): cv.use_id(KeyCollector),
        }
    ),
    synchronous=True,
)
async def enable_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])
    return var


@automation.register_action(
    "key_collector.disable",
    DisableAction,
    automation.maybe_simple_id(
        {
            cv.GenerateID(): cv.use_id(KeyCollector),
        }
    ),
    synchronous=True,
)
async def disable_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])
    return var



import esphome.codegen as cg
from esphome.components import i2c

from ..bmp3xx_base import CONFIG_SCHEMA_BASE, cv, to_code_base

AUTO_LOAD = ["bmp3xx_base"]
CODEOWNERS = ["@latonita"]
DEPENDENCIES = ["i2c"]

bmp3xx_ns = cg.esphome_ns.namespace("bmp3xx_i2c")

BMP3XXI2CComponent = bmp3xx_ns.class_(
    "BMP3XXI2CComponent", cg.PollingComponent, i2c.I2CDevice
)

CONFIG_SCHEMA = CONFIG_SCHEMA_BASE.extend(
    i2c.i2c_device_schema(default_address=0x77)
).extend({cv.GenerateID(): cv.declare_id(BMP3XXI2CComponent)})


async def to_code(config):
    var = await to_code_base(config)
    await i2c.register_i2c_device(var, config)



import logging

import esphome.codegen as cg
from esphome.components import spi
from esphome.components.spi import CONF_SPI_MODE
import esphome.config_validation as cv

from ..spa06_base import CONFIG_SCHEMA_BASE, to_code_base

AUTO_LOAD = ["spa06_base"]
CODEOWNERS = ["@danielkent-net"]
DEPENDENCIES = ["spi"]

spa06_ns = cg.esphome_ns.namespace("spa06_spi")
SPA06SPIComponent = spa06_ns.class_(
    "SPA06SPIComponent", cg.PollingComponent, spi.SPIDevice
)

_LOGGER = logging.getLogger(__name__)

VALID_SPI_MODES = {3: "MODE3", "3": "MODE3", "MODE3": "MODE3"}


def check_spi_mode(config):
    spi_mode = config.get(CONF_SPI_MODE)
    if spi_mode not in VALID_SPI_MODES:
        raise cv.Invalid("SPA06 only supports SPI mode 3")
    return config


CONFIG_SCHEMA = cv.All(
    CONFIG_SCHEMA_BASE.extend(spi.spi_device_schema(default_mode="mode3")).extend(
        {cv.GenerateID(): cv.declare_id(SPA06SPIComponent)}
    ),
    check_spi_mode,
)


async def to_code(config):
    var = await to_code_base(config)
    await spi.register_spi_device(var, config)



import esphome.codegen as cg
from esphome.components import sensor
import esphome.config_validation as cv
from esphome.const import (
    CONF_BATTERY_LEVEL,
    CONF_BATTERY_VOLTAGE,
    CONF_HUMIDITY,
    CONF_ID,
    CONF_SIGNAL_STRENGTH,
    CONF_TEMPERATURE,
    DEVICE_CLASS_BATTERY,
    DEVICE_CLASS_HUMIDITY,
    DEVICE_CLASS_SIGNAL_STRENGTH,
    DEVICE_CLASS_TEMPERATURE,
    DEVICE_CLASS_VOLTAGE,
    ENTITY_CATEGORY_DIAGNOSTIC,
    STATE_CLASS_MEASUREMENT,
    UNIT_CELSIUS,
    UNIT_DECIBEL_MILLIWATT,
    UNIT_PERCENT,
    UNIT_VOLT,
)

from . import bthome_mithermometer_base_schema, setup_bthome_mithermometer

CODEOWNERS = ["@nagyrobi"]

DEPENDENCIES = ["esp32_ble_tracker"]

CONFIG_SCHEMA = bthome_mithermometer_base_schema(
    {
        cv.Optional(CONF_TEMPERATURE): sensor.sensor_schema(
            unit_of_measurement=UNIT_CELSIUS,
            accuracy_decimals=2,
            device_class=DEVICE_CLASS_TEMPERATURE,
            state_class=STATE_CLASS_MEASUREMENT,
        ),
        cv.Optional(CONF_HUMIDITY): sensor.sensor_schema(
            unit_of_measurement=UNIT_PERCENT,
            accuracy_decimals=2,
            device_class=DEVICE_CLASS_HUMIDITY,
            state_class=STATE_CLASS_MEASUREMENT,
        ),
        cv.Optional(CONF_BATTERY_LEVEL): sensor.sensor_schema(
            unit_of_measurement=UNIT_PERCENT,
            accuracy_decimals=0,
            device_class=DEVICE_CLASS_BATTERY,
            state_class=STATE_CLASS_MEASUREMENT,
            entity_category=ENTITY_CATEGORY_DIAGNOSTIC,
        ),
        cv.Optional(CONF_BATTERY_VOLTAGE): sensor.sensor_schema(
            unit_of_measurement=UNIT_VOLT,
            accuracy_decimals=3,
            device_class=DEVICE_CLASS_VOLTAGE,
            state_class=STATE_CLASS_MEASUREMENT,
            icon="mdi:battery-plus",
            entity_category=ENTITY_CATEGORY_DIAGNOSTIC,
        ),
        cv.Optional(CONF_SIGNAL_STRENGTH): sensor.sensor_schema(
            unit_of_measurement=UNIT_DECIBEL_MILLIWATT,
            accuracy_decimals=0,
            device_class=DEVICE_CLASS_SIGNAL_STRENGTH,
            state_class=STATE_CLASS_MEASUREMENT,
            entity_category=ENTITY_CATEGORY_DIAGNOSTIC,
        ),
    }
)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await setup_bthome_mithermometer(var, config)

    if temp_sens := config.get(CONF_TEMPERATURE):
        sens = await sensor.new_sensor(temp_sens)
        cg.add(var.set_temperature(sens))
    if humi_sens := config.get(CONF_HUMIDITY):
        sens = await sensor.new_sensor(humi_sens)
        cg.add(var.set_humidity(sens))
    if batl_sens := config.get(CONF_BATTERY_LEVEL):
        sens = await sensor.new_sensor(batl_sens)
        cg.add(var.set_battery_level(sens))
    if batv_sens := config.get(CONF_BATTERY_VOLTAGE):
        sens = await sensor.new_sensor(batv_sens)
        cg.add(var.set_battery_voltage(sens))
    if sgnl_sens := config.get(CONF_SIGNAL_STRENGTH):
        sens = await sensor.new_sensor(sgnl_sens)
        cg.add(var.set_signal_strength(sens))



import esphome.codegen as cg
from esphome.components import cover, uart
import esphome.config_validation as cv
from esphome.const import CONF_CLOSE_DURATION, CONF_OPEN_DURATION

he60r_ns = cg.esphome_ns.namespace("he60r")
HE60rCover = he60r_ns.class_("HE60rCover", cover.Cover, cg.Component)

CONFIG_SCHEMA = (
    cover.cover_schema(HE60rCover)
    .extend(uart.UART_DEVICE_SCHEMA)
    .extend(cv.COMPONENT_SCHEMA)
    .extend(
        {
            cv.Optional(
                CONF_OPEN_DURATION, default="15s"
            ): cv.positive_time_period_milliseconds,
            cv.Optional(
                CONF_CLOSE_DURATION, default="15s"
            ): cv.positive_time_period_milliseconds,
        }
    )
)

FINAL_VALIDATE_SCHEMA = uart.final_validate_device_schema(
    "he60r",
    baud_rate=1200,
    require_tx=True,
    require_rx=True,
    data_bits=8,
    parity="EVEN",
    stop_bits=1,
)


async def to_code(config):
    var = await cover.new_cover(config)
    await cg.register_component(var, config)
    await uart.register_uart_device(var, config)

    cg.add(var.set_close_duration(config[CONF_CLOSE_DURATION]))
    cg.add(var.set_open_duration(config[CONF_OPEN_DURATION]))



"""Symbol demangling utilities for memory analysis.

This module provides functions for demangling C++ symbol names using c++filt.
"""

from __future__ import annotations

import logging
import re
import subprocess

from .toolchain import find_tool

_LOGGER = logging.getLogger(__name__)

# GCC global constructor/destructor prefix annotations
GCC_PREFIX_ANNOTATIONS = {
    "_GLOBAL__sub_I_": "global constructor for",
    "_GLOBAL__sub_D_": "global destructor for",
}

# GCC optimization suffix pattern (e.g., $isra$0, $part$1, $constprop$2)
GCC_OPTIMIZATION_SUFFIX_PATTERN = re.compile(r"(\$(?:isra|part|constprop)\$\d+)")


def _strip_gcc_annotations(symbol: str) -> tuple[str, str]:
    """Strip GCC optimization suffixes and prefixes from a symbol.

    Args:
        symbol: The mangled symbol name

    Returns:
        Tuple of (stripped_symbol, removed_prefix)
    """
    # Remove GCC optimization markers
    stripped = GCC_OPTIMIZATION_SUFFIX_PATTERN.sub("", symbol)

    # Handle GCC global constructor/initializer prefixes
    prefix = ""
    for gcc_prefix in GCC_PREFIX_ANNOTATIONS:
        if stripped.startswith(gcc_prefix):
            prefix = gcc_prefix
            stripped = stripped[len(prefix) :]
            break

    return stripped, prefix


def _restore_symbol_prefix(prefix: str, stripped: str, demangled: str) -> str:
    """Restore prefix that was removed before demangling.

    Args:
        prefix: Prefix that was removed (e.g., "_GLOBAL__sub_I_")
        stripped: Stripped symbol name
        demangled: Demangled symbol name

    Returns:
        Demangled name with prefix restored/annotated
    """
    if not prefix:
        return demangled

    # Successfully demangled - add descriptive prefix
    if demangled != stripped and (annotation := GCC_PREFIX_ANNOTATIONS.get(prefix)):
        return f"[{annotation}: {demangled}]"

    # Failed to demangle - restore original prefix
    return prefix + demangled


def _restore_symbol_suffix(original: str, demangled: str) -> str:
    """Restore GCC optimization suffix that was removed before demangling.

    Args:
        original: Original symbol name with suffix
        demangled: Demangled symbol name without suffix

    Returns:
        Demangled name with suffix annotation
    """
    if suffix_match := GCC_OPTIMIZATION_SUFFIX_PATTERN.search(original):
        return f"{demangled} [{suffix_match.group(1)}]"
    return demangled


def batch_demangle(
    symbols: list[str],
    cppfilt_path: str | None = None,
    objdump_path: str | None = None,
) -> dict[str, str]:
    """Batch demangle C++ symbol names.

    Args:
        symbols: List of symbol names to demangle
        cppfilt_path: Path to c++filt binary (auto-detected if not provided)
        objdump_path: Path to objdump binary to derive c++filt path from

    Returns:
        Dictionary mapping original symbol names to demangled names
    """
    cache: dict[str, str] = {}

    if not symbols:
        return cache

    # Find c++filt tool
    cppfilt_cmd = cppfilt_path or find_tool("c++filt", objdump_path)
    if not cppfilt_cmd:
        _LOGGER.warning("Could not find c++filt, symbols will not be demangled")
        return {s: s for s in symbols}

    _LOGGER.debug("Demangling %d symbols using %s", len(symbols), cppfilt_cmd)

    # Strip GCC optimization suffixes and prefixes before demangling
    symbols_stripped: list[str] = []
    symbols_prefixes: list[str] = []
    for symbol in symbols:
        stripped, prefix = _strip_gcc_annotations(symbol)
        symbols_stripped.append(stripped)
        symbols_prefixes.append(prefix)

    try:
        result = subprocess.run(
            [cppfilt_cmd],
            input="\n".join(symbols_stripped),
            capture_output=True,
            text=True,
            check=False,
        )
    except (subprocess.SubprocessError, OSError, UnicodeDecodeError) as e:
        _LOGGER.warning("Failed to batch demangle symbols: %s", e)
        return {s: s for s in symbols}

    if result.returncode != 0:
        _LOGGER.warning(
            "c++filt exited with code %d: %s",
            result.returncode,
            result.stderr[:200] if result.stderr else "(no error output)",
        )
        return {s: s for s in symbols}

    # Process demangled output
    demangled_lines = result.stdout.strip().split("\n")

    # Check for output length mismatch
    if len(demangled_lines) != len(symbols):
        _LOGGER.warning(
            "c++filt output mismatch: expected %d lines, got %d",
            len(symbols),
            len(demangled_lines),
        )
        return {s: s for s in symbols}

    failed_count = 0

    for original, stripped, prefix, demangled in zip(
        symbols, symbols_stripped, symbols_prefixes, demangled_lines, strict=True
    ):
        # Add back any prefix that was removed
        demangled = _restore_symbol_prefix(prefix, stripped, demangled)

        # If we stripped a suffix, add it back to the demangled name for clarity
        if original != stripped and not prefix:
            demangled = _restore_symbol_suffix(original, demangled)

        cache[original] = demangled

        # Count symbols that failed to demangle
        if stripped == demangled and stripped.startswith("_Z"):
            failed_count += 1
            if failed_count <= 5:
                _LOGGER.debug("Failed to demangle: %s", original)

    if failed_count > 0:
        _LOGGER.debug(
            "Failed to demangle %d/%d symbols using %s",
            failed_count,
            len(symbols),
            cppfilt_cmd,
        )

    return cache



# Various constants used in MIPI DBI communication
# Various configuration constants for MIPI displays
# Various utility functions for MIPI DBI configuration

from collections.abc import Callable
import functools
from typing import Any, Self

import voluptuous as vol

from esphome.components.const import CONF_COLOR_DEPTH
from esphome.components.display import CONF_SHOW_TEST_CARD, display_ns
import esphome.config_validation as cv
from esphome.const import (
    CONF_BRIGHTNESS,
    CONF_COLOR_ORDER,
    CONF_DIMENSIONS,
    CONF_DISABLED,
    CONF_HEIGHT,
    CONF_INIT_SEQUENCE,
    CONF_INVERT_COLORS,
    CONF_LAMBDA,
    CONF_MIRROR_X,
    CONF_MIRROR_Y,
    CONF_MODEL,
    CONF_OFFSET_HEIGHT,
    CONF_OFFSET_WIDTH,
    CONF_PAGES,
    CONF_RESET_PIN,
    CONF_ROTATION,
    CONF_SWAP_XY,
    CONF_TRANSFORM,
    CONF_WIDTH,
)
from esphome.core import CORE, TimePeriod
from esphome.schema_extractors import SCHEMA_EXTRACT, schema_extractor

LOGGER = cv.logging.getLogger(__name__)

CONF_TRANSFORMS = "transforms"

# All axis transforms a model may support, in the order they appear in the schema.
ALL_TRANSFORMS = (CONF_MIRROR_X, CONF_MIRROR_Y, CONF_SWAP_XY)

ColorOrder = display_ns.enum("ColorMode")

NOP = 0x00
SWRESET = 0x01
RDDID = 0x04
RDDST = 0x09
RDMODE = 0x0A
RDMADCTL = 0x0B
RDPIXFMT = 0x0C
RDIMGFMT = 0x0D
RDSELFDIAG = 0x0F
SLEEP_IN = 0x10
SLPIN = 0x10
SLEEP_OUT = 0x11
SLPOUT = 0x11
PTLON = 0x12
NORON = 0x13
INVERT_OFF = 0x20
INVOFF = 0x20
INVERT_ON = 0x21
INVON = 0x21
ALL_ON = 0x23
WRAM = 0x24
GAMMASET = 0x26
MIPI = 0x26
DISPOFF = 0x28
DISPON = 0x29
CASET = 0x2A
PASET = 0x2B
RASET = 0x2B
RAMWR = 0x2C
WDATA = 0x2C
RAMRD = 0x2E
PTLAR = 0x30
VSCRDEF = 0x33
TEON = 0x35
MADCTL = 0x36
MADCTL_CMD = 0x36
VSCRSADD = 0x37
IDMOFF = 0x38
IDMON = 0x39
COLMOD = 0x3A
PIXFMT = 0x3A
GETSCANLINE = 0x45
BRIGHTNESS = 0x51
WRDISBV = 0x51
RDDISBV = 0x52
WRCTRLD = 0x53
WCE = 0x58
SWIRE1 = 0x5A
SWIRE2 = 0x5B
IFMODE = 0xB0
FRMCTR1 = 0xB1
FRMCTR2 = 0xB2
FRMCTR3 = 0xB3
INVCTR = 0xB4
DFUNCTR = 0xB6
ETMOD = 0xB7
PWCTR1 = 0xC0
PWCTR2 = 0xC1
PWCTR3 = 0xC2
PWCTR4 = 0xC3
PWCTR5 = 0xC4
SPIMODESEL = 0xC4
VMCTR1 = 0xC5
IFCTR = 0xC6
VMCTR2 = 0xC7
GMCTR = 0xC8
SETEXTC = 0xC8
PWSET = 0xD0
VMCTR = 0xD1
PWSETN = 0xD2
RDID4 = 0xD3
RDINDEX = 0xD9
RDID1 = 0xDA
RDID2 = 0xDB
RDID3 = 0xDC
RDIDX = 0xDD
GMCTRP1 = 0xE0
GMCTRN1 = 0xE1
CSCON = 0xF0
PWCTR6 = 0xF6
ADJCTL3 = 0xF7
PAGESEL = 0xFE
PAGESEL1 = 0xFF

MADCTL_MY = 0x80  # Bit 7 Bottom to top
MADCTL_MX = 0x40  # Bit 6 Right to left
MADCTL_MV = 0x20  # Bit 5 Reverse Mode
MADCTL_ML = 0x10  # Bit 4 LCD refresh Bottom to top
MADCTL_RGB = 0x00  # Bit 3 Red-Green-Blue pixel order
MADCTL_BGR = 0x08  # Bit 3 Blue-Green-Red pixel order
MADCTL_MH = 0x04  # Bit 2 LCD refresh right to left

# These bits are used instead of the above bits on some chips, where using MX and MY results in incorrect
# partial updates.
MADCTL_XFLIP = 0x02  # Mirror the display horizontally
MADCTL_YFLIP = 0x01  # Mirror the display vertically

MADCTL_FLIP_FLAG = 0x100  # meta-flag to indicate use of axis flips

# Special constant for delays in command sequences
DELAY_FLAG = 0xFFF  # Special flag to indicate a delay

CONF_PAD_HEIGHT = "pad_height"
CONF_PAD_WIDTH = "pad_width"
CONF_PIXEL_MODE = "pixel_mode"
CONF_USE_AXIS_FLIPS = "use_axis_flips"

PIXEL_MODE_24BIT = "24bit"
PIXEL_MODE_18BIT = "18bit"
PIXEL_MODE_16BIT = "16bit"

PIXEL_MODES = {
    PIXEL_MODE_16BIT: 0x55,
    PIXEL_MODE_18BIT: 0x66,
    PIXEL_MODE_24BIT: 0x77,
}

MODE_RGB = "RGB"
MODE_BGR = "BGR"
COLOR_ORDERS = {
    MODE_RGB: ColorOrder.COLOR_ORDER_RGB,
    MODE_BGR: ColorOrder.COLOR_ORDER_BGR,
}

CONF_HSYNC_BACK_PORCH = "hsync_back_porch"
CONF_HSYNC_FRONT_PORCH = "hsync_front_porch"
CONF_HSYNC_PULSE_WIDTH = "hsync_pulse_width"
CONF_VSYNC_BACK_PORCH = "vsync_back_porch"
CONF_VSYNC_FRONT_PORCH = "vsync_front_porch"
CONF_VSYNC_PULSE_WIDTH = "vsync_pulse_width"
CONF_PCLK_FREQUENCY = "pclk_frequency"
CONF_PCLK_INVERTED = "pclk_inverted"
CONF_NATIVE_HEIGHT = "native_height"
CONF_NATIVE_WIDTH = "native_width"

CONF_DE_PIN = "de_pin"
CONF_PCLK_PIN = "pclk_pin"


def power_of_two(value):
    value = cv.int_range(1, 128)(value)
    if value & (value - 1) != 0:
        raise cv.Invalid("value must be a power of two")
    return value


def validate_dimension(rounding):
    def validator(value):
        value = cv.positive_int(value)
        if value % rounding != 0:
            raise cv.Invalid(f"Dimensions and offsets must be divisible by {rounding}")
        return value

    return validator


def dimension_schema(rounding):
    return cv.Any(
        cv.dimensions,
        cv.Schema(
            {
                cv.Required(CONF_WIDTH): validate_dimension(rounding),
                cv.Required(CONF_HEIGHT): validate_dimension(rounding),
                cv.Optional(CONF_OFFSET_HEIGHT, default=0): validate_dimension(
                    rounding
                ),
                cv.Optional(CONF_OFFSET_WIDTH, default=0): validate_dimension(rounding),
                cv.Optional(CONF_PAD_WIDTH): validate_dimension(rounding),
                cv.Optional(CONF_PAD_HEIGHT): validate_dimension(rounding),
            }
        ),
    )


def map_sequence(value):
    """
    Maps one entry in a sequence to a command and data bytes.
    The format is a repeated sequence of [CMD, <data>] where <data> is s a sequence of bytes. The length is inferred
    from the length of the sequence and should not be explicit.
    A single integer can be provided where there are no data bytes, in which case it is treated as a command.
    A delay can be inserted by specifying "- delay N" where N is in ms
    """
    if isinstance(value, str) and value.lower().startswith("delay "):
        value = value.lower()[6:]
        delay_value = cv.All(
            cv.positive_time_period_milliseconds,
            cv.Range(TimePeriod(milliseconds=1), TimePeriod(milliseconds=255)),
        )(value)
        return DELAY_FLAG, delay_value.total_milliseconds
    value = cv.All(cv.ensure_list(cv.int_range(0, 255)), cv.Length(1, 254))(value)
    return tuple(value)


def flatten_sequence(sequence: tuple | list):
    """
    Flatten an init sequence into a single list of bytes.
    :param sequence:  The list of tuples
    :return: a list of bytes
    """
    return sum(
        tuple(
            (x[1], 0xFF) if x[0] == DELAY_FLAG else (x[0], len(x) - 1) + x[1:]
            for x in sequence
        ),
        (),
    )


def delay(ms):
    return DELAY_FLAG, ms


# Generic placeholder model present in every DriverChip registry; skipped when
# choosing a representative model for schema extraction.
_CUSTOM_MODEL = "CUSTOM"


def model_schema_extractor(
    models: dict[str, Any],
    model_schema: Callable[[dict[str, Any]], Any],
    extra: dict[str, Any] | None = None,
) -> Callable[[Callable[[Any], Any]], Callable[[Any], Any]]:
    """
    Decorate a model-driven display CONFIG_SCHEMA so the language-schema dumper
    can extract it.

    The schema is generated per ``model`` at validation time, so the static
    dumper has nothing to walk. When the dumper passes SCHEMA_EXTRACT, resolve a
    representative schema for a real model (the generic "CUSTOM" placeholder
    over-constrains fields like init_sequence) plus any *extra* keys the model
    needs, e.g. a bus mode, and hand that back; runtime validation is untouched.
    """

    def decorate(config_schema: Callable[[Any], Any]) -> Callable[[Any], Any]:
        @schema_extractor("schema")
        @functools.wraps(config_schema)
        def wrapper(config: Any) -> Any:
            if config is not SCHEMA_EXTRACT:
                return config_schema(config)
            names = sorted(models)
            representative = next((n for n in names if n != _CUSTOM_MODEL), names[0])
            schema = model_schema({CONF_MODEL: representative, **(extra or {})})
            if isinstance(schema, vol.All):
                schema = next(
                    (v for v in schema.validators if isinstance(v, vol.Schema)),
                    schema,
                )
            if isinstance(schema, vol.Schema):
                # The resolved schema pins ``model`` to the representative; expose
                # the full model list so the dumped enum offers every model.
                schema = schema.extend(
                    {cv.Required(CONF_MODEL): cv.one_of(*names, upper=True)}
                )
            return schema

        return wrapper

    return decorate


class DriverChip:
    """
    A class representing a MIPI DBI driver chip model.
    The parameters supplied as defaults will be used to provide default values for the display configuration.
    Pass a ``transforms`` set to restrict which axis transforms (mirror_x, mirror_y, swap_xy) the model
    supports; by default all three are available.
    """

    models: dict[str, Self] = {}

    def __init__(
        self,
        name: str,
        initsequence=None,
        **defaults,
    ):
        name = name.upper()
        self.name = name
        self.initsequence = initsequence
        if CONF_NATIVE_WIDTH in defaults:
            if CONF_WIDTH not in defaults:
                defaults[CONF_WIDTH] = (
                    defaults[CONF_NATIVE_WIDTH]
                    - defaults.get(CONF_OFFSET_WIDTH, 0)
                    - defaults.get(CONF_PAD_WIDTH, 0)
                )
            elif defaults[CONF_WIDTH] > defaults[CONF_NATIVE_WIDTH]:
                defaults[CONF_NATIVE_WIDTH] = defaults[CONF_WIDTH]

        else:
            native_width = (
                defaults.get(CONF_WIDTH, 0)
                + defaults.get(CONF_OFFSET_WIDTH, 0)
                + defaults.get(CONF_PAD_WIDTH, 0)
            )
            if native_width != 0:
                defaults[CONF_NATIVE_WIDTH] = native_width
        if CONF_NATIVE_HEIGHT in defaults:
            if CONF_HEIGHT not in defaults:
                defaults[CONF_HEIGHT] = (
                    defaults[CONF_NATIVE_HEIGHT]
                    - defaults.get(CONF_OFFSET_HEIGHT, 0)
                    - defaults.get(CONF_PAD_HEIGHT, 0)
                )
            elif defaults[CONF_HEIGHT] > defaults[CONF_NATIVE_HEIGHT]:
                defaults[CONF_NATIVE_HEIGHT] = defaults[CONF_HEIGHT]
        else:
            native_height = (
                defaults.get(CONF_HEIGHT, 0)
                + defaults.get(CONF_OFFSET_HEIGHT, 0)
                + defaults.get(CONF_PAD_HEIGHT, 0)
            )
            if native_height != 0:
                defaults[CONF_NATIVE_HEIGHT] = native_height
        self.defaults = defaults
        DriverChip.models[name] = self

    @classmethod
    def get_models(cls):
        """
        Return the current set of models and reset the models dictionary.
        """
        models = cls.models
        cls.models = {}
        return models

    def extend(self, name, **kwargs) -> "DriverChip":
        """
        Extend the current model with additional parameters or a modified init sequence.
        Parameters supplied here will override the defaults of the current model.
        if the initsequence is not provided, the current model's initsequence will be used.
        If add_init_sequence is provided, it will be appended to the current initsequence.
        :param name:
        :param kwargs:
        :return:
        """
        initsequence = list(kwargs.pop("initsequence", self.initsequence))
        initsequence.extend(kwargs.pop("add_init_sequence", ()))
        defaults = self.defaults.copy()
        defaults.update(kwargs)
        return self.__class__(name, initsequence=tuple(initsequence), **defaults)

    def get_default(self, key, fallback: Any = False) -> Any:
        return self.defaults.get(key, fallback)

    @property
    def transforms(self) -> set[str]:
        """
        Return the available transforms for this model.
        """
        if (transforms := self.get_default(CONF_TRANSFORMS, None)) is not None:
            return transforms
        if self.get_default("no_transform", False):
            return set()
        if self.get_default(CONF_SWAP_XY) != cv.UNDEFINED:
            return {CONF_MIRROR_X, CONF_MIRROR_Y, CONF_SWAP_XY}
        raise ValueError(
            "Setting 'swap_xy' to 'cv.UNDEFINED' is no longer supported; set 'transforms' instead"
        )

    def has_hardware_transform(self, config) -> bool:
        """
        Check if the model supports hardware transforms for the given configuration.
        """
        return config.get(CONF_TRANSFORM) != CONF_DISABLED and self.transforms == {
            CONF_MIRROR_X,
            CONF_MIRROR_Y,
            CONF_SWAP_XY,
        }

    def option(self, name, fallback=False) -> cv.Optional:
        return cv.Optional(name, default=self.get_default(name, fallback))

    def rotation_as_transform(self, config) -> bool:
        """
        Check if a rotation can be implemented in hardware using the MADCTL register.
        A rotation of 180 is always possible if x and y mirroring are supported, 90 and 270 are possible if the model supports swapping X and Y.
        """
        if config.get(CONF_TRANSFORM) == CONF_DISABLED:
            return False
        transforms = self.transforms
        rotation = config.get(CONF_ROTATION, 0)
        if rotation == 0 or not transforms:
            return False
        if rotation == 180:
            return CONF_MIRROR_X in transforms and CONF_MIRROR_Y in transforms
        if rotation == 90:
            return CONF_SWAP_XY in transforms and CONF_MIRROR_X in transforms
        return CONF_SWAP_XY in transforms and CONF_MIRROR_Y in transforms

    def get_dimensions(
        self, config, swap: bool = True
    ) -> tuple[int, int, int, int, int, int]:
        """
        Return the dimensions of the current model.
        :param config: The current configuration
        :param swap: If width/height should be swapped when axes are swapped.
        :return: A tuple (width, height, offset_width, offset_height, pad_width, pad_height).
        """

        transform = self.get_transform(config)
        if CONF_DIMENSIONS in config:
            # Explicit dimensions, just use as is
            dimensions = config[CONF_DIMENSIONS]
            if isinstance(dimensions, dict):
                native_width = self.get_default(CONF_NATIVE_WIDTH, 0)
                native_height = self.get_default(CONF_NATIVE_HEIGHT, 0)
                if transform.get(CONF_SWAP_XY) is True:
                    native_width, native_height = native_height, native_width
                width = dimensions[CONF_WIDTH]
                height = dimensions[CONF_HEIGHT]
                offset_width = dimensions[CONF_OFFSET_WIDTH]
                offset_height = dimensions[CONF_OFFSET_HEIGHT]
                if CONF_PAD_WIDTH in dimensions:
                    pad_width = dimensions[CONF_PAD_WIDTH]
                    native_width = width + offset_width + pad_width
                elif native_width == 0:
                    pad_width = 0
                    native_width = width + offset_width
                else:
                    pad_width = native_width - width - offset_width
                if CONF_PAD_HEIGHT in dimensions:
                    pad_height = dimensions[CONF_PAD_HEIGHT]
                    native_height = height + offset_height + pad_height
                elif native_height == 0:
                    pad_height = 0
                    native_height = height + offset_height
                else:
                    pad_height = native_height - height - offset_height
                if (
                    pad_width + offset_width >= native_width
                    or pad_height + offset_height >= native_height
                ):
                    raise cv.Invalid("Dimensions exceed native size", [CONF_DIMENSIONS])
                if pad_width < 0 or pad_height < 0:
                    raise cv.Invalid("Invalid offsets", [CONF_DIMENSIONS])

                return width, height, offset_width, offset_height, pad_width, pad_height

            # Must be a tuple
            width, height = dimensions
            return width, height, 0, 0, 0, 0

        # Default dimensions, use model defaults

        width = self.get_default(CONF_WIDTH)
        height = self.get_default(CONF_HEIGHT)
        native_width = self.get_default(CONF_NATIVE_WIDTH, 0)
        native_height = self.get_default(CONF_NATIVE_HEIGHT, 0)
        offset_width = self.get_default(CONF_OFFSET_WIDTH, 0)
        offset_height = self.get_default(CONF_OFFSET_HEIGHT, 0)
        pad_width = self.get_default(
            CONF_PAD_WIDTH, native_width - width - offset_width
        )
        pad_height = self.get_default(
            CONF_PAD_HEIGHT, native_height - height - offset_height
        )

        if pad_width < 0 or pad_height < 0:
            raise cv.Invalid("Offsets exceed native size", [CONF_DIMENSIONS])

        # if mirroring axes and there are offsets, also mirror the offsets to cater for situations where
        # the offset is asymmetric
        if transform.get(CONF_MIRROR_X):
            offset_width, pad_width = pad_width, offset_width
        if transform.get(CONF_MIRROR_Y):
            offset_height, pad_height = pad_height, offset_height
        # Swap default dimensions if swap_xy is set, or if rotation is 90/270, and we are not using a buffer
        if swap and transform.get(CONF_SWAP_XY) is True:
            width, height = height, width
            offset_height, offset_width = offset_width, offset_height
            pad_width, pad_height = pad_height, pad_width
        return width, height, offset_width, offset_height, pad_width, pad_height

    def get_base_transform(self, config):
        transform = config.get(
            CONF_TRANSFORM,
            {
                CONF_MIRROR_X: self.get_default(CONF_MIRROR_X),
                CONF_MIRROR_Y: self.get_default(CONF_MIRROR_Y),
                CONF_SWAP_XY: self.get_default(CONF_SWAP_XY),
            },
        )
        if isinstance(transform, dict):
            return transform

        # Transform is disabled
        return {
            CONF_MIRROR_X: False,
            CONF_MIRROR_Y: False,
            CONF_SWAP_XY: False,
            CONF_TRANSFORM: False,
        }

    def get_transform(self, config) -> dict[str, bool]:
        transform = self.get_base_transform(config)
        # Can we use the MADCTL register to set the rotation?
        transform[CONF_TRANSFORM] = self.rotation_as_transform(config)
        return transform

    def transform_schema(self):
        """
        Build the schema for the ``transform`` config option of this model.

        Each transform the model supports is a required boolean. A transform the model does not
        support may be omitted or set to ``false``; setting it to ``true`` reports a clear error
        naming the unsupported transform instead of a generic "extra keys not allowed".
        """
        supported = self.transforms

        def unsupported(name):
            def validator(value):
                if cv.boolean(value):
                    raise cv.Invalid(f"'{name}' is not supported by this model")
                return False

            return validator

        schema = {}
        for name in ALL_TRANSFORMS:
            if name in supported:
                schema[cv.Required(name)] = cv.boolean
            else:
                schema[cv.Optional(name, default=False)] = unsupported(name)
        return cv.Any(cv.Schema(schema), cv.one_of(CONF_DISABLED, lower=True))

    def get_madctl(self, transform: dict, config: dict) -> int:
        """
        Convert a transform to MADCTL bits
        :param transform: The transform dict
        :param use_flip: Whether to use axis flips
        :return: MADCTL value
        """
        use_flip = config.get(CONF_USE_AXIS_FLIPS, False)
        madctl = MADCTL_FLIP_FLAG if use_flip else 0
        if transform[CONF_MIRROR_X]:
            madctl |= MADCTL_XFLIP if use_flip else MADCTL_MX
        if transform[CONF_MIRROR_Y]:
            madctl |= MADCTL_YFLIP if use_flip else MADCTL_MY
        if transform.get(CONF_SWAP_XY) is True:  # Exclude Undefined
            madctl |= MADCTL_MV
        if config[CONF_COLOR_ORDER] == MODE_BGR:
            madctl |= MADCTL_BGR
        return madctl

    def add_madctl(self, sequence: list, config: dict):
        # Add the MADCTL command to the sequence based on the base configuration.
        # Rotation is not applied here, it will be done at runtime.
        transform = self.get_transform(config)
        madctl = self.get_madctl(transform, config)
        sequence.append((MADCTL, madctl & 0xFF))

    def skip_command(self, command: str):
        """
        Allow suppressing a standard command in the init sequence.
        """
        return self.get_default(f"no_{command.lower()}", False)

    def get_sequence(self, config, add_madctl=True, add_reset=False) -> tuple[int, ...]:
        """
        Create the init sequence for the display.
        Use the default sequence from the model, if any, and append any custom sequence provided in the config.
        Append SLPOUT (if not already in the sequence) and DISPON to the end of the sequence
        MADCTL will be set if add_madctl is True
        If add_reset is True, a reset is prepended: a software reset when no reset pin
        is configured (and the model doesn't skip it), followed by a settling delay that
        both a software and a hardware reset require.
        Returns the init sequence
        """
        sequence = list(self.initsequence or ())
        custom_sequence = config.get(CONF_INIT_SEQUENCE, [])
        sequence.extend(custom_sequence)
        # Ensure each command is a tuple
        sequence = [x if isinstance(x, tuple) else (x,) for x in sequence]

        if add_reset:
            reset: list = []
            # A software reset is only needed when there is no hardware reset pin.
            if CONF_RESET_PIN not in config and not self.skip_command("SWRESET"):
                reset.append((SWRESET,))
            # Both a software and a hardware reset need a settling delay before further commands.
            reset.append(delay(10))
            sequence = reset + sequence

        # Set pixel format if not already in the custom sequence
        pixel_mode = config[CONF_PIXEL_MODE]
        if not isinstance(pixel_mode, int):
            if not pixel_mode.endswith("bit"):
                pixel_mode = f"{pixel_mode}bit"
            pixel_mode = PIXEL_MODES[pixel_mode]
        sequence.append((PIXFMT, pixel_mode))

        if self.rotation_as_transform(config):
            LOGGER.info("Using hardware transform to implement rotation")
        if add_madctl:
            self.add_madctl(sequence, config)
        if config[CONF_INVERT_COLORS]:
            sequence.append((INVON,))
        else:
            sequence.append((INVOFF,))
        if brightness := config.get(CONF_BRIGHTNESS, self.get_default(CONF_BRIGHTNESS)):
            sequence.append((BRIGHTNESS, brightness))
        # Add a SLPOUT command if required.
        if not self.skip_command("SLPOUT"):
            # A zero delay will delay until 120ms after reset
            sequence.append(delay(0))
            sequence.append((SLPOUT,))
            sequence.append(delay(10))
        sequence.append((DISPON,))
        # Add a delay here because additional commands may be added after this at runtime.
        sequence.append(delay(10))

        # Flatten the sequence into a list of bytes, with the length of each command
        # or the delay flag inserted where needed
        return flatten_sequence(sequence)

    def check_requirements(self) -> None:
        """
        Raise a friendly error if any component this model requires is not configured.

        This runs during schema validation (before ID references are resolved) so that a
        model whose default pins live on a pin expander reports the missing expander clearly
        instead of a cryptic "Couldn't find ID" from the unresolved pin reference.

        Also logs a warning if the model is deprecated.
        """
        if deprecation_reason := self.get_default("deprecation_reason"):
            LOGGER.warning(
                "Display model %s is deprecated: %s", self.name, deprecation_reason
            )
        if requirements := self.get_default("requires", set()):
            # ``raw_config`` is populated before any component schema runs during a real
            # validation, so presence of a required component is simply a top-level key.
            # When it is absent (e.g. a unit test that invokes the schema directly) there
            # is no config to check against, so skip.
            global_config = CORE.raw_config
            if global_config is None:
                return
            missing = {x for x in requirements if x not in global_config}
            if missing:
                reqstr = ", ".join(f"'{x}'" for x in sorted(missing))
                raise cv.Invalid(
                    f"{self.name} requires component{'s' if len(missing) > 1 else ''} {reqstr} to be configured"
                )


def requires_buffer(config) -> bool:
    """
    Check if the display configuration requires a buffer. It will do so if any drawing methods are configured.
    :param config:
    :return:  True if a buffer is required, False otherwise
    """
    return any(
        config.get(key) for key in (CONF_LAMBDA, CONF_PAGES, CONF_SHOW_TEST_CARD)
    )


def get_color_depth(config) -> int:
    """
    Get the color depth in bits from the configuration.
    """
    return int(config[CONF_COLOR_DEPTH].removesuffix("bit"))



"""Subprocess entry point that applies ESPHome's PlatformIO patches.

Invoked via ``python -m esphome.platformio.runner`` instead of
``python -m platformio`` so that the patches (incremental rebuild
preservation, download retries) apply inside the subprocess. Running
PlatformIO in a subprocess keeps its ``sys.path`` mutations and other
global state from leaking into the ESPHome process.
"""

from __future__ import annotations

import logging
from pathlib import Path
import sys
import time
from typing import Any

_LOGGER = logging.getLogger(__name__)


def patch_structhash() -> None:
    """Avoid full rebuilds when files are added or removed.

    PlatformIO clears the build dir whenever its structure hash changes.
    We replace that with an mtime check against ``platformio.ini`` so
    incremental builds are preserved unless the project config changed.
    """
    from platformio.run import cli, helpers

    def patched_clean_build_dir(build_dir, *_args):
        from platformio import fs
        from platformio.project.helpers import get_project_dir

        platformio_ini = Path(get_project_dir()) / "platformio.ini"
        build_dir = Path(build_dir)

        if (
            build_dir.is_dir()
            and platformio_ini.stat().st_mtime > build_dir.stat().st_mtime
        ):
            fs.rmtree(build_dir)

        if not build_dir.is_dir():
            build_dir.mkdir(parents=True)

    helpers.clean_build_dir = patched_clean_build_dir
    cli.clean_build_dir = patched_clean_build_dir


def patch_file_downloader() -> None:
    """Retry PlatformIO package downloads with exponential backoff.

    PlatformIO's ``FileDownloader`` uses an ``HTTPSession`` without built-in
    retry. We wrap ``__init__`` to retry on transient failures and close the
    session between attempts so a new TCP connection can route to a different
    CDN edge node. We catch both ``PackageException`` (raised when the server
    returns a non-200 status such as 502/503) and ``OSError`` -- which covers
    ``requests.exceptions.ConnectionError``, ``ReadTimeout``, and
    ``ChunkedEncodingError`` (all subclasses of ``OSError``) that get raised
    when the connection is aborted before a response is parsed.
    """
    from platformio.package.download import FileDownloader
    from platformio.package.exception import PackageException

    if getattr(FileDownloader.__init__, "_esphome_patched", False):
        return

    original_init = FileDownloader.__init__

    def patched_init(self, *args: Any, **kwargs: Any) -> None:
        max_retries = 5

        for attempt in range(max_retries):
            try:
                original_init(self, *args, **kwargs)
                return
            except (PackageException, OSError) as e:
                if attempt < max_retries - 1:
                    delay = 2 ** (attempt + 1)
                    _LOGGER.warning(
                        "Package download failed: %s. "
                        "Retrying in %d seconds... (attempt %d/%d)",
                        str(e),
                        delay,
                        attempt + 1,
                        max_retries,
                    )
                    # pylint: disable=protected-access,broad-except
                    try:
                        if (
                            hasattr(self, "_http_response")
                            and self._http_response is not None
                        ):
                            self._http_response.close()
                        if hasattr(self, "_http_session"):
                            self._http_session.close()
                    except Exception:  # noqa: BLE001
                        pass
                    # pylint: enable=protected-access,broad-except
                    time.sleep(delay)
                else:
                    raise

    patched_init._esphome_patched = True  # type: ignore[attr-defined]  # pylint: disable=protected-access
    FileDownloader.__init__ = patched_init


_IGNORE_LIB_WARNINGS = "(?:Hash|Update)"
# Regex patterns matched against each line of PlatformIO output. Lines that
# match are dropped by RedirectText before they reach the parent process.
# Patterns are anchored at the start of the line (RedirectText uses
# ``re.match``). Disabled when the user passes ``-v`` / ``--verbose`` to
# ``esphome compile``.
FILTER_PLATFORMIO_LINES = [
    r"Verbose mode can be enabled via `-v, --verbose` option.*",
    r"CONFIGURATION: https://docs.platformio.org/.*",
    r"DEBUG: Current.*",
    r"LDF Modes:.*",
    r"LDF: Library Dependency Finder -> https://bit.ly/configure-pio-ldf.*",
    f"Looking for {_IGNORE_LIB_WARNINGS} library in registry",
    f"Warning! Library `.*'{_IGNORE_LIB_WARNINGS}.*` has not been found in PlatformIO Registry.",
    f"You can ignore this message, if `.*{_IGNORE_LIB_WARNINGS}.*` is a built-in library.*",
    r"Scanning dependencies...",
    r"Found \d+ compatible libraries",
    r"Memory Usage -> https://bit.ly/pio-memory-usage",
    r"Found: https://platformio.org/lib/show/.*",
    r"Using cache: .*",
    r"Installing dependencies",
    r"Library Manager: Already installed, built-in library",
    r"Building in .* mode",
    r"Advanced Memory Usage is available via .*",
    r"Merged .* ELF section",
    r"esptool.py v.*",
    r"esptool v.*",
    r"Checking size .*",
    r"Retrieving maximum program size .*",
    r"PLATFORM: .*",
    r"PACKAGES:.*",
    r" - framework-arduinoespressif.* \(.*\)",
    r" - tool-esptool.* \(.*\)",
    r" - toolchain-.* \(.*\)",
    r"Creating BIN file .*",
    r"Warning! Could not find file \".*.crt\"",
    r"Warning! Arduino framework as an ESP-IDF component doesn't handle the `variant` field! The default `esp32` variant will be used.",
    r"Warning: DEPRECATED: 'esptool.py' is deprecated. Please use 'esptool' instead. The '.py' suffix will be removed in a future major release.",
    r"Warning: esp-idf-size exited with code 2",
    r"esp_idf_size: error: unrecognized arguments: --ng",
    r"Package configuration completed successfully",
]


def main() -> int:
    patch_structhash()
    patch_file_downloader()

    # Wrap stdout/stderr with RedirectText before PlatformIO runs:
    #
    # 1. RedirectText.isatty() unconditionally returns True. Click, tqdm, and
    #    PlatformIO's own progress-bar code check ``stream.isatty()`` to
    #    decide whether to emit TTY-format output (``\r`` cursor moves, ANSI
    #    colors, fancy progress bars). With the wrapper in place they always
    #    emit TTY format, even when our real stdout is a pipe to the parent
    #    process. Downstream consumers (local terminals and the Home
    #    Assistant dashboard log viewer) render the TTY control sequences
    #    correctly, so the user sees real progress bars.
    #
    # 2. FILTER_PLATFORMIO_LINES is applied inside RedirectText.write() in
    #    this subprocess, so noisy PlatformIO output is dropped before it
    #    ever leaves the runner. This replaces the parent-side filtering
    #    that was lost when we switched from in-process to subprocess — the
    #    parent's ``subprocess.run`` uses ``.fileno()`` on RedirectText and
    #    bypasses its ``write()`` path entirely.
    #
    # Filtering is disabled when the user passed -v / --verbose to
    # ``esphome compile``, preserving the previous in-process behavior where
    # verbose mode let all PlatformIO output through unfiltered.
    from esphome.util import RedirectText

    is_verbose = any(arg in ("-v", "--verbose") for arg in sys.argv[1:])
    filter_lines = None if is_verbose else FILTER_PLATFORMIO_LINES

    sys.stdout = RedirectText(sys.stdout, filter_lines=filter_lines)
    sys.stderr = RedirectText(sys.stderr, filter_lines=filter_lines)

    import platformio.__main__

    return platformio.__main__.main() or 0


if __name__ == "__main__":
    sys.exit(main())



"""ESP-IDF direct build API for ESPHome."""

from dataclasses import dataclass, field
import json
import logging
import os
from pathlib import Path
import re
import shutil
import subprocess

from esphome.components.esp32.const import KEY_ESP32, KEY_FLASH_SIZE, KEY_IDF_VERSION
from esphome.const import (
    CONF_COMPILE_PROCESS_LIMIT,
    CONF_ESPHOME,
    CONF_FRAMEWORK,
    CONF_SOURCE,
)
from esphome.core import CORE, EsphomeError
from esphome.espidf.framework import check_esp_idf_install, get_framework_env
from esphome.espidf.size_summary import print_summary
from esphome.helpers import add_git_ceiling_directory

_LOGGER = logging.getLogger(__name__)

DOMAIN = "espidf_toolchain"


@dataclass
class _CacheData:
    paths: dict[str, tuple] = field(default_factory=dict)
    env: dict[str, dict[str, str]] = field(default_factory=dict)
    cmake_output: dict[Path, str] = field(default_factory=dict)
    cmake_tools: dict[Path, dict[str, Path]] = field(default_factory=dict)


def _cache() -> _CacheData:
    if DOMAIN not in CORE.data:
        CORE.data[DOMAIN] = _CacheData()
    return CORE.data[DOMAIN]


def _get_core_framework_version():
    return str(CORE.data[KEY_ESP32][KEY_IDF_VERSION])


def _get_framework_source_override() -> str | None:
    """Return the user-supplied esp32.framework.source override, if any.

    The override lets a user point the IDF tarball download at a custom URL
    (mirror, fork, local server). Substitutions like ``{VERSION}`` /
    ``{MAJOR}`` etc. work the same as in the default mirror list.
    """
    if CORE.config is None:
        return None
    return CORE.config.get(KEY_ESP32, {}).get(CONF_FRAMEWORK, {}).get(CONF_SOURCE)


def _get_esphome_esp_idf_paths(
    version: str | None = None,
) -> tuple[os.PathLike, os.PathLike]:
    version = version or _get_core_framework_version()
    paths = _cache().paths
    if version not in paths:
        paths[version] = check_esp_idf_install(
            version, source_url=_get_framework_source_override()
        )
    return paths[version]


def _get_idf_path(version: str | None = None) -> Path | None:
    """Get IDF_PATH from environment or common locations."""
    # Use provided IDF framework if available
    if "IDF_PATH" in os.environ:
        return Path(os.environ["IDF_PATH"])
    return Path(_get_esphome_esp_idf_paths(version)[0])


def _get_idf_env(version: str | None = None) -> dict[str, str]:
    """Get environment variables needed for ESP-IDF build."""
    version = version or _get_core_framework_version()
    env_cache = _cache().env
    if version not in env_cache:
        env_cache[version] = os.environ.copy()

        # Use provided IDF framework if available
        if "IDF_PATH" not in os.environ:
            env_cache[version] |= get_framework_env(
                *_get_esphome_esp_idf_paths(version)
            )

        # Cap git's repo search at the config directory so ESP-IDF's
        # `git describe` for the app version can't error out on an
        # uninitialized or corrupt git repo in a parent directory.
        add_git_ceiling_directory(env_cache[version], CORE.config_dir)
    return env_cache[version]


def _get_cmake_output(build_dir) -> str:
    cmake_output_cache = _cache().cmake_output
    if build_dir not in cmake_output_cache:
        cmd = ["cmake", "-LA", "-N", "."]

        env = _get_idf_env()
        result = subprocess.run(
            cmd,
            cwd=build_dir,
            env=env,
            capture_output=True,
            text=True,
            check=False,
        )

        if result.returncode != 0:
            raise RuntimeError(f"CMake failed: {result.stderr}")

        cmake_output_cache[build_dir] = result.stdout
    return cmake_output_cache[build_dir]


def _get_cmake_tool_path(var_name: str) -> Path:
    build_dir = CORE.relative_build_path("build")
    cmake_output = _get_cmake_output(build_dir)

    cmake_tools_cache = _cache().cmake_tools
    if build_dir not in cmake_tools_cache:
        cmake_tools_cache[build_dir] = {}

    if var_name not in cmake_tools_cache[build_dir]:
        pattern = rf"^{var_name}:FILEPATH=(.+)$"
        match = re.search(pattern, cmake_output, re.MULTILINE)

        if not match:
            raise RuntimeError(f"{var_name} not found in CMake output")

        path = match.group(1).strip()
        cmake_tools_cache[build_dir][var_name] = Path(path)

    return cmake_tools_cache[build_dir][var_name]


def _get_idf_tool(name: str) -> str:
    """Return the path to an executable from the ESP-IDF environment PATH or raise if not found."""
    env = _get_idf_env()
    executable = shutil.which(name, path=env.get("PATH", None))
    if executable is None:
        raise EsphomeError(
            f"{name} executable not found in ESP-IDF environment. "
            "Check that the IDF environment is correctly set up."
        )
    return executable


def run_idf_py(
    *args,
    cwd: Path | None = None,
    capture_output: bool = False,
    jobs: int | None = None,
) -> int | str:
    """Run idf.py with the given arguments."""
    idf_path = _get_idf_path()
    if idf_path is None:
        raise EsphomeError("ESP-IDF not found")

    env = _get_idf_env()
    if jobs is not None:
        env = {**env, "IDF_PY_BUILD_JOBS": str(jobs)}
    python_executable = _get_idf_tool("python")
    idf_py = idf_path / "tools" / "idf.py"
    # Dispatch idf.py through esphome.espidf.runner, which wraps
    # sys.stdout/sys.stderr so ``isatty()`` reports True. This keeps CMake,
    # Ninja, and idf.py's own progress-bar code emitting TTY-format output
    # (``\r`` cursor moves, ANSI colors, fancy progress bars) even when our
    # real stdout is a pipe — e.g. when esphome is running under the Home
    # Assistant dashboard add-on. The runner is a plain script (not a
    # ``python -m`` module) because IDF's Python venv does not have the
    # esphome package installed.
    runner_py = Path(__file__).parent / "runner.py"

    cmd = [python_executable, str(runner_py), str(idf_py)] + list(args)

    if cwd is None:
        cwd = CORE.build_path

    _LOGGER.debug("Running: %s", " ".join(cmd))
    _LOGGER.debug("  in directory: %s", cwd)

    if capture_output:
        result = subprocess.run(
            cmd,
            cwd=cwd,
            env=env,
            capture_output=True,
            text=True,
            check=False,
        )
        if result.returncode != 0:
            _LOGGER.error("idf.py failed:\n%s", result.stderr)
        return result.stdout
    result = subprocess.run(
        cmd,
        cwd=cwd,
        env=env,
        check=False,
    )
    return result.returncode


def _get_sdkconfig_args() -> list[str]:
    """Get cmake -D flags for the sdkconfig file, if it exists."""
    sdkconfig_path = CORE.relative_build_path(f"sdkconfig.{CORE.name}")
    if sdkconfig_path.is_file():
        return ["-D", f"SDKCONFIG={sdkconfig_path}"]
    return []


def run_reconfigure() -> int:
    """Run cmake reconfigure only (no build)."""
    return run_idf_py(*_get_sdkconfig_args(), "reconfigure")


def has_outdated_files():
    """Check if the build configuration is stale.

    Returns True if required build files are missing or if ESPHome's
    resolved build inputs are newer than CMakeCache.txt:

    - ``sdkconfig.<name>.esphomeinternal`` -- the canonical "what state
      did ESPHome resolve the YAML to" snapshot. Any change in build
      flags, enabled components, framework version, or target ends up
      rewriting it (we embed a ``# ESPHOME_IDF_VERSION=`` comment line
      for the version case where the option set would otherwise be
      identical).
    - ``src/idf_component.yml`` -- the project manifest. Managed
      component additions/removals (e.g. via ``add_idf_component``) can
      happen without any sdkconfig impact, and ``_write_idf_component_yml``
      already deletes ``dependencies.lock`` on a change but that signal
      gets lost as soon as the lock is missing.

    We deliberately don't watch:
    - The top-level/src ``CMakeLists.txt`` -- ESPHome owns those, and
      ninja already tracks them as configure-time deps. Including them
      causes a perpetual reconfigure loop because CMake doesn't restamp
      ``CMakeCache.txt`` when only ``idf_build_set_property`` values
      change between configures.
    - ``$IDF_PATH`` and CMake's ``build/config/`` -- both have mtime
      semantics that fire after the wrong configure (or not at all in
      common cases like in-place IDF version replacement). The sdkconfig
      and manifest hashes subsume the meaningful signal.
    """
    cmakecache_txt_path = CORE.relative_build_path("build/CMakeCache.txt")
    build_config_path = CORE.relative_build_path("build/config")
    sdkconfig_internal_path = CORE.relative_build_path(
        f"sdkconfig.{CORE.name}.esphomeinternal"
    )
    idf_component_yml_path = CORE.relative_build_path("src/idf_component.yml")
    dependency_lock_path = CORE.relative_build_path("dependencies.lock")
    build_ninja_path = CORE.relative_build_path("build/build.ninja")

    if not build_config_path.is_dir() or not any(build_config_path.iterdir()):
        return True
    if not cmakecache_txt_path.is_file():
        return True
    if not build_ninja_path.is_file():
        return True
    if (
        dependency_lock_path.is_file()
        and dependency_lock_path.stat().st_mtime > build_ninja_path.stat().st_mtime
    ):
        return True

    cmakecache_txt_mtime = cmakecache_txt_path.stat().st_mtime
    return any(
        f.stat().st_mtime > cmakecache_txt_mtime
        for f in [sdkconfig_internal_path, idf_component_yml_path]
        if f.exists()
    )


def need_reconfigure() -> bool:
    from esphome.build_gen.espidf import has_discovered_components

    # We need to reconfigure either if the files are outdated or if there is no component discovered
    return has_outdated_files() or not has_discovered_components()


def _patch_memory_segments():
    """Patch memory.ld to expand IRAM/DRAM for testing mode.

    Mirrors the PlatformIO iram_fix.py.script logic for native IDF builds.
    Must be called after cmake configure (which generates memory.ld) and
    before the build/link step.
    """
    # Same sizes as iram_fix.py.script
    testing_iram_size = 0x200000  # 2MB
    testing_dram_size = 0x200000  # 2MB

    memory_ld = CORE.relative_build_path(
        "build", "esp-idf", "esp_system", "ld", "memory.ld"
    )
    if not memory_ld.is_file():
        _LOGGER.warning("Could not find linker script at %s", memory_ld)
        return

    content = memory_ld.read_text()
    patches = []

    def _patch_segment(text, segment_name, new_size):
        pattern = rf"({re.escape(segment_name)}\s*\([^)]*\)\s*:\s*org\s*=\s*.+?,\s*len\s*=\s*)(\S+[^\n]*)"
        if match := re.search(pattern, text, re.DOTALL):
            replacement = f"{match.group(1)}{new_size:#x}"
            new_text = text[: match.start()] + replacement + text[match.end() :]
            if new_text != text:
                return new_text, True
        return text, False

    content, patched = _patch_segment(content, "iram0_0_seg", testing_iram_size)
    if patched:
        patches.append(f"IRAM={testing_iram_size:#x}")

    content, patched = _patch_segment(content, "dram0_0_seg", testing_dram_size)
    if patched:
        patches.append(f"DRAM={testing_dram_size:#x}")

    if patches:
        memory_ld.write_text(content)
        _LOGGER.info("Patched %s in %s for testing mode", ", ".join(patches), memory_ld)
    else:
        _LOGGER.warning("Could not patch memory segments in %s", memory_ld)


def run_compile(config, verbose: bool) -> int:
    """Compile the ESP-IDF project.

    Uses two-phase configure to auto-discover available components:
    1. If no previous build, configure with minimal REQUIRES to discover components
    2. Regenerate CMakeLists.txt with discovered components
    3. Run full build
    """
    from esphome.build_gen.espidf import write_project

    # Check if we need to do discovery phase
    if need_reconfigure():
        _LOGGER.info("Discovering available ESP-IDF components...")
        write_project(minimal=True)
        rc = run_reconfigure()
        if rc != 0:
            _LOGGER.error("Component discovery failed")
            return rc
        _LOGGER.info("Regenerating CMakeLists.txt with discovered components...")
        write_project(minimal=False)
        if CORE.testing_mode:
            # Reconfigure again so cmake is up to date with the full
            # component list before the build's idf.py invocation runs --
            # idf.py build would otherwise re-run cmake and regenerate
            # memory.ld, wiping the DRAM/IRAM patches applied below.
            # Outside testing mode ninja's own configure-time dep on
            # CMakeLists.txt handles the re-run as part of the build step.
            rc = run_reconfigure()
            if rc != 0:
                _LOGGER.error("Reconfigure with discovered components failed")
                return rc

    # In testing mode, generate the linker script first, patch DRAM/IRAM sizes,
    # then build. memory.ld is regenerated by ninja during the build phase,
    # so we must patch after it's generated but before linking (same timing
    # as iram_fix.py.script's AddPreAction hook in the PlatformIO path).
    if CORE.testing_mode:
        memory_ld = CORE.relative_build_path(
            "build", "esp-idf", "esp_system", "ld", "memory.ld"
        )
        build_dir = CORE.relative_build_path("build")
        # Build just the memory.ld target - ninja needs the path relative to build dir
        memory_ld_target = os.path.relpath(str(memory_ld), str(build_dir))
        env = _get_idf_env()
        ninja_executable = _get_idf_tool("ninja")
        result = subprocess.run(
            [ninja_executable, "-C", str(build_dir), memory_ld_target],
            env=env,
            check=False,
        )
        if result.returncode != 0:
            _LOGGER.error("Failed to generate linker script")
            return result.returncode
        _patch_memory_segments()

    # Build
    args = []

    if verbose:
        args.append("-v")

    args.extend(_get_sdkconfig_args())
    args.append("build")
    args.append("size")

    rc = run_idf_py(*args, jobs=config[CONF_ESPHOME].get(CONF_COMPILE_PROCESS_LIMIT))
    if rc == 0:
        size_json = CORE.relative_build_path("build", "esp_idf_size.json")
        partitions = CORE.relative_build_path("partitions.csv")
        print_summary(size_json, partitions if partitions.is_file() else None)
    return rc


def get_firmware_path() -> Path:
    """Get the path to the compiled firmware binary.

    This is the file idf.py writes directly (named after the project),
    not the copy used for OTA/factory downloads below.
    """
    build_dir = CORE.relative_build_path("build")
    return build_dir / f"{CORE.name}.bin"


def get_factory_firmware_path() -> Path:
    """Get the path to the factory firmware (with bootloader).

    Uses the PlatformIO ``firmware.factory.bin`` naming convention so
    the dashboard's download handler — which requests files by name
    relative to ``firmware_bin_path.parent`` — finds it. Without this,
    the native IDF path produced ``<name>.factory.bin`` and the
    dashboard returned 500 trying to locate ``firmware.factory.bin``.
    """
    build_dir = CORE.relative_build_path("build")
    return build_dir / "firmware.factory.bin"


def get_ota_firmware_path() -> Path:
    """Get the path to the OTA firmware binary.

    Uses the PlatformIO ``firmware.ota.bin`` naming convention for the
    same dashboard-compatibility reason as ``get_factory_firmware_path``.
    """
    build_dir = CORE.relative_build_path("build")
    return build_dir / "firmware.ota.bin"


def get_elf_path() -> Path:
    """Get the path to the firmware ELF file.

    idf.py writes ``<build>/<name>.elf`` directly; this returns the
    ``<build>/firmware.elf`` copy created by ``create_elf_copy`` so
    the dashboard's "download ELF" link can find it under the
    PlatformIO-convention name.
    """
    build_dir = CORE.relative_build_path("build")
    return build_dir / "firmware.elf"


def get_objdump_path() -> Path:
    return _get_cmake_tool_path("CMAKE_OBJDUMP")


def get_readelf_path() -> Path:
    return _get_cmake_tool_path("CMAKE_READELF")


def get_addr2line_path() -> Path:
    return _get_cmake_tool_path("CMAKE_ADDR2LINE")


def get_idedata() -> dict | None:
    """Derive idedata from the build's compile_commands.json.

    The native ESP-IDF toolchain has no ``pio run -t idedata`` equivalent, but
    its CMake build emits ``build/compile_commands.json``. Parse that into the
    idedata fields IDE integrations and clang-tidy expect, cached alongside the
    PlatformIO idedata path. Returns None if the compile DB doesn't exist yet.
    """
    from esphome.espidf.idedata import idedata_from_build

    compile_commands = CORE.relative_build_path("build", "compile_commands.json")
    if not compile_commands.is_file():
        _LOGGER.debug("No %s yet; skipping idedata generation", compile_commands)
        return None

    cache = CORE.relative_internal_path("idedata", f"{CORE.name}.json")
    if cache.is_file() and cache.stat().st_mtime >= compile_commands.stat().st_mtime:
        try:
            return json.loads(cache.read_text(encoding="utf-8"))
        except ValueError:
            pass

    data = idedata_from_build(compile_commands)
    data["prog_path"] = str(get_elf_path())
    cache.parent.mkdir(parents=True, exist_ok=True)
    cache.write_text(json.dumps(data, indent=2) + "\n", encoding="utf-8")
    return data


def create_factory_bin() -> bool:
    """Create factory.bin by merging bootloader, partition table, and app."""
    build_dir = CORE.relative_build_path("build")
    flasher_args_path = build_dir / "flasher_args.json"

    if not flasher_args_path.is_file():
        _LOGGER.warning("flasher_args.json not found, cannot create factory.bin")
        return False

    try:
        with flasher_args_path.open(encoding="utf-8") as f:
            flash_data = json.load(f)
    except (json.JSONDecodeError, OSError) as e:
        _LOGGER.error("Failed to read flasher_args.json: %s", e)
        return False

    # Get flash size from config
    flash_size = CORE.data[KEY_ESP32][KEY_FLASH_SIZE]

    # Build esptool merge command
    sections = []
    for addr, fname in sorted(
        flash_data.get("flash_files", {}).items(), key=lambda kv: int(kv[0], 16)
    ):
        file_path = build_dir / fname
        if file_path.is_file():
            sections.extend([addr, str(file_path)])
        else:
            _LOGGER.warning("Flash file not found: %s", file_path)

    if not sections:
        _LOGGER.warning("No flash sections found")
        return False

    output_path = get_factory_firmware_path()
    chip = flash_data.get("extra_esptool_args", {}).get("chip", "esp32")

    env = _get_idf_env()
    python_executable = _get_idf_tool("python")
    cmd = [
        python_executable,
        "-m",
        "esptool",
        "--chip",
        chip,
        "merge_bin",
        "--flash_size",
        flash_size,
        "--output",
        str(output_path),
    ] + sections

    _LOGGER.info("Creating factory.bin...")
    result = subprocess.run(cmd, env=env, capture_output=True, text=True, check=False)

    if result.returncode != 0:
        _LOGGER.error("Failed to create factory.bin: %s", result.stderr)
        return False

    _LOGGER.info("Created: %s", output_path)
    return True


def create_ota_bin() -> bool:
    """Copy the firmware to firmware.ota.bin for ESPHome OTA compatibility."""
    firmware_path = get_firmware_path()
    ota_path = get_ota_firmware_path()

    if not firmware_path.is_file():
        _LOGGER.warning("Firmware not found: %s", firmware_path)
        return False

    shutil.copy(firmware_path, ota_path)
    _LOGGER.info("Created: %s", ota_path)
    return True


def create_elf_copy() -> bool:
    """Copy the ELF binary to firmware.elf for dashboard compatibility.

    idf.py writes the ELF at ``<build>/<name>.elf``; the dashboard's
    "download ELF" link requests the literal filename ``firmware.elf``
    (PlatformIO convention), so copy it to that name.
    """
    build_dir = CORE.relative_build_path("build")
    src_elf = build_dir / f"{CORE.name}.elf"
    dst_elf = get_elf_path()

    if not src_elf.is_file():
        _LOGGER.warning("ELF not found: %s", src_elf)
        return False

    shutil.copy(src_elf, dst_elf)
    _LOGGER.info("Created: %s", dst_elf)
    return True



import esphome.codegen as cg
from esphome.components import i2c, output
import esphome.config_validation as cv
from esphome.const import CONF_CHANNEL, CONF_ID

from .. import CONF_GP8403_ID, GP8403Component, gp8403_ns

DEPENDENCIES = ["gp8403"]

GP8403Output = gp8403_ns.class_(
    "GP8403Output", cg.Component, i2c.I2CDevice, output.FloatOutput
)

CONFIG_SCHEMA = output.FLOAT_OUTPUT_SCHEMA.extend(
    {
        cv.GenerateID(): cv.declare_id(GP8403Output),
        cv.GenerateID(CONF_GP8403_ID): cv.use_id(GP8403Component),
        cv.Required(CONF_CHANNEL): cv.int_range(min=0, max=1),
    }
).extend(cv.COMPONENT_SCHEMA)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)
    await output.register_output(var, config)

    await cg.register_parented(var, config[CONF_GP8403_ID])

    cg.add(var.set_channel(config[CONF_CHANNEL]))



import esphome.codegen as cg
from esphome.components import binary_sensor, sensor
import esphome.config_validation as cv
from esphome.const import CONF_SENSOR_ID, CONF_THRESHOLD

analog_threshold_ns = cg.esphome_ns.namespace("analog_threshold")

AnalogThresholdBinarySensor = analog_threshold_ns.class_(
    "AnalogThresholdBinarySensor", binary_sensor.BinarySensor, cg.Component
)

CONF_UPPER = "upper"
CONF_LOWER = "lower"

CONFIG_SCHEMA = (
    binary_sensor.binary_sensor_schema(AnalogThresholdBinarySensor)
    .extend(
        {
            cv.Required(CONF_SENSOR_ID): cv.use_id(sensor.Sensor),
            cv.Required(CONF_THRESHOLD): cv.Any(
                cv.templatable(cv.float_),
                cv.Schema(
                    {
                        cv.Required(CONF_UPPER): cv.templatable(cv.float_),
                        cv.Required(CONF_LOWER): cv.templatable(cv.float_),
                    }
                ),
            ),
        }
    )
    .extend(cv.COMPONENT_SCHEMA)
)


async def to_code(config):
    var = await binary_sensor.new_binary_sensor(config)
    await cg.register_component(var, config)

    sens = await cg.get_variable(config[CONF_SENSOR_ID])
    cg.add(var.set_sensor(sens))

    if isinstance(config[CONF_THRESHOLD], dict):
        lower = await cg.templatable(config[CONF_THRESHOLD][CONF_LOWER], [], cg.float_)
        upper = await cg.templatable(config[CONF_THRESHOLD][CONF_UPPER], [], cg.float_)
    else:
        lower = await cg.templatable(config[CONF_THRESHOLD], [], cg.float_)
        upper = lower
    cg.add(var.set_upper_threshold(upper))
    cg.add(var.set_lower_threshold(lower))



from esphome import automation
import esphome.codegen as cg
from esphome.components import i2c
import esphome.config_validation as cv
from esphome.const import (
    CONF_ADDRESS,
    CONF_CHANNEL,
    CONF_DIRECTION,
    CONF_ID,
    CONF_SPEED,
)

DEPENDENCIES = ["i2c"]

CODEOWNERS = ["@max246"]

MULTI_CONF = True

grove_tb6612fng_ns = cg.esphome_ns.namespace("grove_tb6612fng")
GROVE_TB6612FNG = grove_tb6612fng_ns.class_(
    "GroveMotorDriveTB6612FNG", cg.Component, i2c.I2CDevice
)
GROVETB6612FNGMotorRunAction = grove_tb6612fng_ns.class_(
    "GROVETB6612FNGMotorRunAction", automation.Action
)
GROVETB6612FNGMotorBrakeAction = grove_tb6612fng_ns.class_(
    "GROVETB6612FNGMotorBrakeAction", automation.Action
)
GROVETB6612FNGMotorStopAction = grove_tb6612fng_ns.class_(
    "GROVETB6612FNGMotorStopAction", automation.Action
)
GROVETB6612FNGMotorStandbyAction = grove_tb6612fng_ns.class_(
    "GROVETB6612FNGMotorStandbyAction", automation.Action
)
GROVETB6612FNGMotorNoStandbyAction = grove_tb6612fng_ns.class_(
    "GROVETB6612FNGMotorNoStandbyAction", automation.Action
)
GROVETB6612FNGMotorChangeAddressAction = grove_tb6612fng_ns.class_(
    "GROVETB6612FNGMotorChangeAddressAction", automation.Action
)

DIRECTION_TYPE = {
    "FORWARD": 1,
    "BACKWARD": 2,
}

CONFIG_SCHEMA = (
    cv.Schema(
        {
            cv.Required(CONF_ID): cv.declare_id(GROVE_TB6612FNG),
        }
    )
    .extend(cv.COMPONENT_SCHEMA)
    .extend(i2c.i2c_device_schema(0x14))
)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)
    await i2c.register_i2c_device(var, config)


@automation.register_action(
    "grove_tb6612fng.run",
    GROVETB6612FNGMotorRunAction,
    cv.Schema(
        {
            cv.Required(CONF_ID): cv.use_id(GROVE_TB6612FNG),
            cv.Required(CONF_CHANNEL): cv.templatable(cv.int_range(min=0, max=1)),
            cv.Required(CONF_SPEED): cv.templatable(cv.int_range(min=0, max=255)),
            cv.Required(CONF_DIRECTION): cv.enum(DIRECTION_TYPE, upper=True),
        }
    ),
    synchronous=True,
)
async def grove_tb6612fng_run_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])

    template_channel = await cg.templatable(config[CONF_CHANNEL], args, cg.uint8)
    template_speed = await cg.templatable(config[CONF_SPEED], args, cg.uint16)
    cg.add(var.set_channel(template_channel))
    cg.add(var.set_speed(template_speed))
    cg.add(var.set_direction(config[CONF_DIRECTION] == "FORWARD"))
    return var


@automation.register_action(
    "grove_tb6612fng.break",
    GROVETB6612FNGMotorBrakeAction,
    cv.Schema(
        {
            cv.Required(CONF_ID): cv.use_id(GROVE_TB6612FNG),
            cv.Required(CONF_CHANNEL): cv.templatable(cv.int_range(min=0, max=1)),
        }
    ),
    synchronous=True,
)
async def grove_tb6612fng_break_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])

    template_channel = await cg.templatable(config[CONF_CHANNEL], args, cg.uint8)
    cg.add(var.set_channel(template_channel))
    return var


@automation.register_action(
    "grove_tb6612fng.stop",
    GROVETB6612FNGMotorStopAction,
    cv.Schema(
        {
            cv.Required(CONF_ID): cv.use_id(GROVE_TB6612FNG),
            cv.Required(CONF_CHANNEL): cv.templatable(cv.int_range(min=0, max=1)),
        }
    ),
    synchronous=True,
)
async def grove_tb6612fng_stop_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])

    template_channel = await cg.templatable(config[CONF_CHANNEL], args, cg.uint8)
    cg.add(var.set_channel(template_channel))
    return var


@automation.register_action(
    "grove_tb6612fng.standby",
    GROVETB6612FNGMotorStandbyAction,
    cv.Schema(
        {
            cv.Required(CONF_ID): cv.use_id(GROVE_TB6612FNG),
        }
    ),
    synchronous=True,
)
async def grove_tb6612fng_standby_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])

    return var


@automation.register_action(
    "grove_tb6612fng.no_standby",
    GROVETB6612FNGMotorNoStandbyAction,
    cv.Schema(
        {
            cv.Required(CONF_ID): cv.use_id(GROVE_TB6612FNG),
        }
    ),
    synchronous=True,
)
async def grove_tb6612fng_no_standby_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])

    return var


@automation.register_action(
    "grove_tb6612fng.change_address",
    GROVETB6612FNGMotorChangeAddressAction,
    cv.Schema(
        {
            cv.Required(CONF_ID): cv.use_id(GROVE_TB6612FNG),
            cv.Required(CONF_ADDRESS): cv.i2c_address,
        }
    ),
    synchronous=True,
)
async def grove_tb6612fng_change_address_to_code(config, action_id, template_arg, args):
    var = cg.new_Pvariable(action_id, template_arg)
    await cg.register_parented(var, config[CONF_ID])

    template_channel = await cg.templatable(config[CONF_ADDRESS], args, cg.uint8)
    cg.add(var.set_address(template_channel))
    return var



import esphome.codegen as cg
from esphome.components import spi
import esphome.config_validation as cv

from ..bme280_base import CONFIG_SCHEMA_BASE, to_code_base

AUTO_LOAD = ["bme280_base"]
CODEOWNERS = ["@apbodrov"]
DEPENDENCIES = ["spi"]


bme280_spi_ns = cg.esphome_ns.namespace("bme280_spi")
BME280SPIComponent = bme280_spi_ns.class_(
    "BME280SPIComponent", cg.PollingComponent, spi.SPIDevice
)

CONFIG_SCHEMA = CONFIG_SCHEMA_BASE.extend(spi.spi_device_schema()).extend(
    {cv.GenerateID(): cv.declare_id(BME280SPIComponent)}
)


async def to_code(config):
    var = await to_code_base(config)
    await spi.register_spi_device(var, config)



import re

from esphome import config_validation as cv
from esphome.const import CONF_ARGS, CONF_FORMAT

CONF_IF_NAN = "if_nan"


f_regex = re.compile(
    r"""
    (                                   # start of capture group 1
    %                                   # literal "%"
    [-+0 #]{0,5}                        # optional flags
    (?:\d+|\*)?                         # width
    (?:\.(?:\d+|\*))?                   # precision
    (?:hh|h|ll|l|j|z|t|L|w|I|I32|I64)?  # size
    f                                   # type
    )
    """,
    flags=re.VERBOSE,
)
c_regex = re.compile(
    r"""
    (                                   # start of capture group 1
    %                                   # literal "%"
    [-+0 #]{0,5}                        # optional flags
    (?:\d+|\*)?                         # width
    (?:\.(?:\d+|\*))?                   # precision
    (?:hh|h|ll|l|j|z|t|L|w|I|I32|I64)?  # size
    [cCdiouxXeEfgGaAnpsSZ]              # type
    )
    """,
    flags=re.VERBOSE,
)


def validate_printf(value):
    format_string = value[CONF_FORMAT]
    matches = c_regex.findall(format_string)
    if len(matches) != len(value[CONF_ARGS]):
        raise cv.Invalid(
            f"Found {len(matches)} printf-patterns ({', '.join(matches)}), but {len(value[CONF_ARGS])} args were given!"
        )

    if value.get(CONF_IF_NAN) and len(f_regex.findall(format_string)) != 1:
        raise cv.Invalid(
            "Use of 'if_nan' requires a single valid printf-pattern of type %f"
        )
    return value
